import time
import uuid
from abc import ABC, abstractmethod
from math import ceil
from pathlib import Path

import numpy as np
import pandas as pd
from rich.progress import (
    BarColumn,
    MofNCompleteColumn,
    Progress,
    TextColumn,
    TimeElapsedColumn,
)
from rich.table import Table
from textdistance import levenshtein

from socialgene.base.socialgene import SocialGene
from socialgene.clustermap.serialize import SerializeToClustermap
from socialgene.compare_gene_clusters.compare_gene_clusters import BGCComparison
from socialgene.neo4j.neo4j import GraphDriver
from socialgene.utils.logging import CONSOLE, log

progress_bar = Progress(
    TextColumn("Pulling target cluster data from database..."),
    BarColumn(),
    MofNCompleteColumn(),
    TextColumn("• Time elapsed "),
    TimeElapsedColumn(),
    transient=True,
)


class BgcComp2:
    def __init__(
        self,
        levenshtein_include_internal_nonortholog,
        levenshtein_only_orthologs,
        percent_of_query,
        jaccard,
        modscore,
    ):
        self.levenshtein_include_internal_nonortholog = (
            levenshtein_include_internal_nonortholog
        )
        self.levenshtein_only_orthologs = levenshtein_only_orthologs
        self.percent_of_query = percent_of_query
        self.jaccard = jaccard
        self.modscore = modscore


def truncate_string(str_input, max_length):
    str_end = ".."
    length = len(str_input)
    if length > max_length:
        return str_input[: max_length - len(str_end)] + str_end
    else:
        return str_input


##################################################################################################################
class SearchBase(ABC):
    """
    Base class for searching for a biosynthetic gene clusters (BGCs) in a SocialGene database

    Args:
        gene_clusters_must_have_x_matches (float): Minimum number of distinct query protein matches required for a BGC to be considered a hit. <1 is a fraction of the number of query proteins, >1 is the number of query proteins.
        assemblies_must_have_x_matches (float): Minimum number of distinct query protein matches for a genome to be considered. <1 is a fraction of the number of query proteins, >1 is the number of query proteins.
        nucleotide_sequences_must_have_x_matches (float): Minimum number of distinct query protein matches required for a nucleotide sequence to be considered. <1 is a fraction of the number of query proteins, >1 is the number of query proteins.
        modscore_cutoff (float, optional): Minimum score required for a BGC to be considered a hit. Defaults to 0.8.
        break_bgc_on_gap_of(int, optional): Putative BGCs are split when matched proteins are separated by n base pairs. Defaults to 20000.
        target_bgc_padding (int, optional): Number of base pairs to add to the start and end of a target BGC. Defaults to 10000.


    Attributes:
        gene_clusters_must_have_x_matches (float): Minimum number of distinct query protein matches required for a BGC to be considered a hit. <1 is a fraction of the number of query proteins, >1 is the number of query proteins.
        assemblies_must_have_x_matches (float): Minimum number of distinct query protein matches for a genome to be considered. <1 is a fraction of the number of query proteins, >1 is the number of query proteins.
        nucleotide_sequences_must_have_x_matches (float): Minimum number of distinct query protein matches required for a nucleotide sequence to be considered. <1 is a fraction of the number of query proteins, >1 is the number of query proteins.
        input_bgc_id (str): ID of the input biosynthetic gene cluster (BGC).
        input_assembly (Assembly): Assembly object of the input BGC.
        modscore_cutoff (float): Minimum score required for a BGC to be considered a hit.
        break_bgc_on_gap_of (int): Number of base pairs to break a BGC on if a gap is greater than this value.
        sg_object (SocialGene): SocialGene object containing the database.
        raw_search_results_df (pd.DataFrame): Initial search results as a pandas DataFrame.
        sorted_bgcs (pd.DataFrame): Order of the BGCs in the search results as a pandas DataFrame.
        link_df (pd.DataFrame): Linkage information between BGCs in the search results as a pandas DataFrame.
        group_df (pd.DataFrame): Grouping information for BGCs in the search results as a pandas DataFrame.
        n_searched_proteins (int): Number of proteins searched in the database.
        target_bgc_padding (int): Number of base pairs to add to the start and end of a target BGC.
        working_search_results_df (pd.DataFrame): Working search results as a pandas DataFrame.

    Methods:
        input (): SocialGene object containing a single BGC to search (optional, must provide this or a gbk_path); or Path to GenBank file containing a single BGC to search (optional, must provide this or a sg_object).
        search(): Abstract method to be implemented in child classes. Searches the database for BGCs.
        _modify_input_bgc_name(): Modifies the name of the input BGC to avoid overriding existing BGCs in the database.
        _input_bgc_info(): Logs information about the input BGC.
        read_sg_object(sg_object: SocialGene): Reads a SocialGene object as the database.
        read_input_bgc(gbk_path: str): Reads an input BGC from a GenBank file.
        process_results(): Processes the search results to filter and group BGCs.
    """

    def __init__(
        self,
        input,
        gene_clusters_must_have_x_matches: float,
        assemblies_must_have_x_matches: float,
        nucleotide_sequences_must_have_x_matches: float,
        modscore_cutoff: float = 0.8,
        target_bgc_padding=10000,
        break_bgc_on_gap_of: int = 20000,
        adaptive_padding: bool = True,
    ) -> None:
        super().__init__()
        self.gene_clusters_must_have_x_matches = gene_clusters_must_have_x_matches
        self.assemblies_must_have_x_matches = assemblies_must_have_x_matches
        self.nucleotide_sequences_must_have_x_matches = (
            nucleotide_sequences_must_have_x_matches
        )
        self.input_bgc_id = None
        self.input_assembly = None
        self.modscore_cutoff = modscore_cutoff
        self.break_bgc_on_gap_of = break_bgc_on_gap_of
        self.sg_object = SocialGene()
        self.raw_search_results_df = pd.DataFrame()
        self.sorted_bgcs = None
        self.link_df = pd.DataFrame()
        self.group_df = pd.DataFrame()
        self.n_searched_proteins = None
        self.target_bgc_padding = target_bgc_padding
        self.search_padding = self.target_bgc_padding
        self.working_search_results_df = pd.DataFrame()
        self.outdegree_df = pd.DataFrame()
        self.bgc_df = pd.DataFrame()
        self.primary_bgc_regions = pd.DataFrame()
        if (
            isinstance(input, SocialGene)
            or isinstance(input, str)
            or isinstance(input, Path)
        ):
            self.read_input_bgc(input)
            self.gbk_path = input
        else:
            raise ValueError("Must provide either sg_object or gbk_path")

        if adaptive_padding:
            input_len = max(i.end for i in self.input_bgc.features) - min(
                i.start for i in self.input_bgc.features
            )
            input_len += 500
            if input_len > self.target_bgc_padding:
                log.error(f"Adaptive padding: {input_len}")
                self.search_padding = input_len
        self.n_searched_proteins = len(self.sg_object.proteins)
        self._compare_two_gene_clusters_score = BgcComp2

    @abstractmethod
    def prioritize_input_proteins(
        self,
        scatter: bool = False,
    ):
        # To be implemented in the child class
        # scatter (bool, optional): Choose a random subset of proteins to search that are spread across the length of the input BGC. Defaults to False.
        ...

    @abstractmethod
    def search(self):
        # To be implemented in the child class
        # This function should set self.raw_search_results_df to a pd.DataFrame:
        # pd.DataFrame(columns=['assembly_uid', 'nucleotide_uid', 'target', 'n_start', 'n_end','query'])
        ...

    @property
    @abstractmethod
    def outdegree_table(self):
        # prints info about the prioritized proteins for the input BGC
        ...

    @abstractmethod
    def _outdegree_table_stats(self):
        # generates table for outdegree_table()
        ...

    def _modify_input_bgc_name(self):
        # modify input BGC assembly name so if it also exists in the DB it won't be overridden
        self.input_bgc_id = list(self.sg_object.assemblies.keys())[0]
        self.modified_input_bgc_name = f"socialgene_query_{self.input_bgc_id}"
        self.sg_object.assemblies[self.modified_input_bgc_name] = (
            self.sg_object.assemblies.pop(self.input_bgc_id)
        )
        self.sg_object.assemblies[self.modified_input_bgc_name].uid = (
            self.modified_input_bgc_name
        )
        self.input_assembly = self.sg_object.assemblies[self.modified_input_bgc_name]

    def create_input_bgcs(self):
        for locus in self.input_assembly.loci.values():
            locus.add_bgcs_by_feature(features=locus.features)
        self.input_bgc = list(self.input_assembly.gene_clusters)[0]

    def _input_bgc_info(self):
        log.info(
            f"Input BGC has {len(self.sg_object.proteins)} proteins and/or pseudogenes"
        )

    def read_input_bgc(self, input: str):
        # parse input BGC
        if isinstance(input, SocialGene):
            self.sg_object = input
        else:
            self.sg_object.parse(input)
        self._modify_input_bgc_name()
        self._input_bgc_info()
        self.create_input_bgcs()

    def filter(self, drop_raw_search_results_df=False):
        if self.raw_search_results_df.empty:
            log.warning("Stopping, no search results to process")
            return
        log.info(
            f"Starting with matches across {self.raw_search_results_df.assembly_uid.nunique()} genomes"
        )
        self.working_search_results_df = self.raw_search_results_df
        if drop_raw_search_results_df:
            del self.raw_search_results_df
        # filter assemblies, nucleotide sequences based on the number of matches to unique input BGC proteins
        self._filter_assemblies(threshold=self.assemblies_must_have_x_matches)
        if self.working_search_results_df.empty:
            ValueError("No hits found after filtering at the assembly level")
        self._filter_nucleotides(
            threshold=self.nucleotide_sequences_must_have_x_matches
        )
        if self.working_search_results_df.empty:
            raise ValueError("No hits found after filtering at the assembly level")

    def _primary_bgc_regions(self):
        return self._collapse_cluster(
            self._filter_clusters(
                df=self.working_search_results_df,
                threshold=self.gene_clusters_must_have_x_matches,
            )
        )

    def annotate(self):
        try:
            self._create_links()
            self._choose_group()
            self._rank_order_bgcs()
        except Exception as e:
            log.warning(e)

    def _compare_two_gene_clusters(self, df):
        target_cluster_min_start = (
            df.target_feature.dropna().apply(lambda x: x.start).min()
        )
        target_cluster_max_end = df.target_feature.dropna().apply(lambda x: x.end).max()
        non_orthologs = [
            i
            for i in df.target_gene_cluster.dropna().iloc[0].features
            if i.feature_is_protein()
            and i.start > target_cluster_min_start
            and i.end < target_cluster_max_end
            and i not in df.target_feature.to_list()
        ]
        # add the non-orthologs to the df dataframe as new rows with the same columns and a random uuid as the query
        non_ortholog_df = pd.DataFrame(
            [
                {
                    "query": np.nan,
                    "target": str(uuid.uuid4()),
                    "target_feature": i,
                    "q_start": np.inf,
                    "t_start": i.start,
                    "strand_y": i.strand,
                }
                for i in non_orthologs
            ]
        )
        if not non_ortholog_df.empty:
            non_ortholog_df["query"] = non_ortholog_df["target"]
            full_t_obj = pd.concat([df, non_ortholog_df], ignore_index=True)
        else:
            full_t_obj = df
        levenshtein_include_internal_nonortholog_for = levenshtein(
            list(
                full_t_obj.dropna(subset="target_feature").sort_values(
                    ["t_start"], ascending=True
                )["query"]
            ),
            list(
                full_t_obj.dropna(subset="query_feature").sort_values(
                    ["q_start"], ascending=True
                )["query"]
            ),
        ) / len(full_t_obj)
        levenshtein_include_internal_nonortholog_rev = levenshtein(
            list(
                full_t_obj.dropna(subset="target_feature").sort_values(
                    ["t_start"], ascending=True
                )["query"]
            ),
            list(
                full_t_obj.dropna(subset="query_feature").sort_values(
                    ["q_start"], ascending=False
                )["query"]
            ),
        ) / len(full_t_obj)
        levenshtein_only_orthologs_for = levenshtein(
            list(df.dropna().sort_values(["t_start"], ascending=True)["query"]),
            list(df.dropna().sort_values(["q_start"], ascending=True)["query"]),
        ) / len(df.dropna())
        levenshtein_only_orthologs_rev = levenshtein(
            list(df.dropna().sort_values(["t_start"], ascending=True)["query"]),
            list(df.dropna().sort_values(["q_start"], ascending=False)["query"]),
        ) / len(df.dropna())

        levenshtein_include_internal_nonortholog = 1 - min(
            levenshtein_include_internal_nonortholog_for,
            levenshtein_include_internal_nonortholog_rev,
        )
        levenshtein_only_orthologs = 1 - min(
            levenshtein_only_orthologs_for, levenshtein_only_orthologs_rev
        )
        # percent_of_query = how many query proteins were seen / total query proteins * 100
        percent_of_query = (
            df.dropna().query_feature.nunique()
            / len(df[~df.query_gene_cluster.isna()].query_gene_cluster.iloc[0].features)
            * 100
        )
        percent_of_query = int(round(percent_of_query, 0))
        query_len = len(
            df[~df.query_gene_cluster.isna()].query_gene_cluster.iloc[0].features
        )
        # this is more complicated than just retrieving all the target proteins because any proteins in the added window must be removed first
        target_len = len(
            [
                i
                for i in df.target_gene_cluster.dropna().iloc[0].features
                if i.feature_is_protein()
                and i.start >= target_cluster_min_start
                and i.end <= target_cluster_max_end
            ]
        )
        intersection = (
            df.groupby("target_gene_cluster")["query_feature"].nunique().iloc[0]
        )
        jac = intersection / (query_len + target_len - intersection)
        modscore = (jac * 2) + levenshtein_include_internal_nonortholog
        return self._compare_two_gene_clusters_score(
            levenshtein_include_internal_nonortholog=levenshtein_include_internal_nonortholog,
            levenshtein_only_orthologs=levenshtein_only_orthologs,
            percent_of_query=percent_of_query,
            jaccard=jac,
            modscore=modscore,
        )

    def _compare_all_gcs_to_input_gc(self):
        # create a 3-column dataframe that represents the input gene cluster's features (proteins)
        q_obj = pd.DataFrame(
            [
                {
                    "query_feature": i,
                    "q_start": i.start,
                    "strand": i.strand,
                }
                for i in list(self.input_bgc.features_sorted_by_midpoint)
            ]
        )
        for k, t_obj in self.link_df.groupby(
            ["query_gene_cluster", "target_gene_cluster"], sort=False
        ):
            score = self._compare_two_gene_clusters(
                pd.merge(
                    q_obj,
                    t_obj,
                    on="query_feature",
                    how="outer",
                    suffixes=("", "_delme"),
                ),
            )
            yield {
                "query_gene_cluster": k[1],
                "target_gene_cluster": k[0],
                "levenshtein_include_internal_nonortholog": score.levenshtein_include_internal_nonortholog,
                "levenshtein_only_orthologs": score.levenshtein_only_orthologs,
                "percent_of_query": score.percent_of_query,
                "jaccard": score.jaccard,
                "modscore": score.modscore,
            }

    def _compare_bgcs_by_jaccard_and_levenshtein(
        self,
    ):
        """Sorts assembly IDs by comparing the levenshtein distance of ordered query proteins (forward and reverse)"""
        return pd.DataFrame(self._compare_all_gcs_to_input_gc()).sort_values(
            by="modscore", ascending=False
        )

    def _compare_bgcs_by_median_bitscore(
        self,
    ):
        temp = self.link_df[["query_gene_cluster", "target_gene_cluster", "score"]]
        return (
            temp.groupby(["query_gene_cluster", "target_gene_cluster"])["score"]
            .median()
            .sort_values(ascending=False)
            .reset_index()
        )

    def _fraction_matched(self):
        df = (
            self.link_df.groupby(["target_gene_cluster"])["query"]
            .nunique()
            .apply(lambda x: x / len(self.input_bgc.proteins))
            .reset_index()
        )
        df.columns = ["gene_cluster", "fraction_matched"]
        return df

    def _rank_order_bgcs(self, threshold):
        """Sorts assemblies by comparing the jaccard and levenshtein distance of ordered
            query proteins (forward and reverse) and breaking ties with the median bitscore of RBH between BGCs

        Args:
            threshold (float): The minimum jaccard score required for a BGC to be considered a hit. Think of this as the number of input bgc proteins a putative bgc must have to be considered a hit. (0 < threshold < 1)
        Returns:
            list[Assembly]: List of SocialGene Assembly objects
        """
        temp = pd.merge(
            self._compare_bgcs_by_jaccard_and_levenshtein(),
            self._compare_bgcs_by_median_bitscore(),
            left_on="query_gene_cluster",
            right_on="target_gene_cluster",
            how="inner",
        )
        temp.drop(
            ["query_gene_cluster_y", "target_gene_cluster_y"], axis=1, inplace=True
        )
        primary = (
            temp.loc[temp["jaccard"] >= threshold]
            .sort_values(by=["modscore", "score"], ascending=False)
            .copy()
        )
        primary.loc[:, "query_gene_cluster_x_assembly"] = (
            primary.query_gene_cluster_x.apply(lambda x: x.parent.parent.uid)
        )
        primary.drop_duplicates(subset=["query_gene_cluster_x_assembly"], inplace=True)
        return primary["query_gene_cluster_x"].to_list()

    def _bgc_regions_to_sg_object(self, df):
        now = time.time()
        log.info(
            f"Pulling data from the database for {df.cluster.nunique()} putative BGCs"
        )
        with progress_bar as pg:
            task = pg.add_task("Adding best hits...", total=len(df))
            for index, df_row in df.iterrows():
                # pull in info from database to sg_object
                _ = self.sg_object.fill_given_locus_range(
                    locus_uid=df_row.loc["nucleotide_uid"],
                    start=df_row.loc["n_start"] - self.search_padding,
                    end=df_row.loc["n_end"] + self.search_padding,
                )
                # add bgc to locus
                pg.update(task, advance=1)
        self.sg_object._drop_all_cross_origin()

        then = time.time()
        log.info(f"Time to fill: {int(then - now)} seconds")

    def _create_link_df(
        self, query_gene_cluster, target_gene_cluster, tool="blastp", **kwargs
    ):
        """read the args, things are backward b/c context is searching the found bgc against the input bgc

        Args:
            query_gene_cluster (_type_): the database putative gene cluster
            target_gene_cluster (_type_): the user input gene cluster

        Returns:
            _type_: _description_
        """

        comparator = BGCComparison(
            tool=tool,
        )
        protein_comparisons_df = comparator.compare(
            query_gene_cluster.protein_iter, target_gene_cluster.protein_iter, **kwargs
        )
        # Turn the protein uid in the blastp etc output into protein objects from the cluster sg_objects
        protein_comparisons_df["query_protein"] = protein_comparisons_df["query"].apply(
            lambda x: query_gene_cluster.proteins[x]
        )
        protein_comparisons_df["target_protein"] = protein_comparisons_df[
            "target"
        ].apply(lambda x: target_gene_cluster.proteins[x])
        # expand to get one row per feature per protein (one non-redunant protein can have multiple features)
        q_obj = pd.DataFrame(
            [
                {
                    "query_gene_cluster": query_gene_cluster,
                    "query_feature": i,
                    "query_protein": i.protein,
                    "q_start": i.start,
                    "strand": i.strand,
                }
                for i in list(query_gene_cluster.features_sorted_by_midpoint)
            ]
        )
        # expand to get one row per feature per protein (one non-redunant protein can have multiple features)
        t_obj = pd.DataFrame(
            [
                {
                    "target_gene_cluster": target_gene_cluster,
                    "target_feature": i,
                    "target_protein": i.protein,
                    "t_start": i.start,
                    "strand": i.strand,
                }
                for i in list(target_gene_cluster.features_sorted_by_midpoint)
            ]
        )
        q_temp = pd.merge(protein_comparisons_df, q_obj, on="query_protein", how="left")
        return (
            pd.merge(q_temp, t_obj, on="target_protein", how="left")
            .sort_values("t_start")
            .rename(columns={comparator.tool.score_column: "score"}, inplace=False)
        )

    def _create_links(self, tool="hmmer", **kwargs) -> pd.DataFrame:
        """
        Loop through gene_cluster compare the proteins to the input BGC

        Returns:
        pd.DataFrame: A DataFrame containing the links between the input BGC and the putative BGCs.
        The DataFrame has the following columns:
            - query: The protein hash of the input BGC protein.
            - target: The protein hash of the putative BGC protein.
            - score: The score of the link between the input BGC protein and the putative BGC protein.
            - assembly_uid: The assembly UID of the putative BGC.
            - nucleotide_uid: The nucleotide UID of the putative BGC.
            - cluster: The cluster of the putative BGC.
        """
        # Initialize an empty DataFrame to store the links
        log.info("Start: Creating links")
        link_df = pd.DataFrame()
        progress_bar = Progress(
            TextColumn("[progress.description]{task.description}"),
            BarColumn(),
            MofNCompleteColumn(),
            TextColumn("• Time elapsed "),
            TimeElapsedColumn(),
            transient=True,
        )
        log.info(
            f"Finding reciprocal best hits; protein similarity via {BGCComparison(tool=tool).tool.name}"
        )
        with progress_bar as pg:
            task = pg.add_task(
                description="Creating links",
                total=len(list(self.sg_object.get_all_gene_clusters())),
            )
            for target_gene_cluster in self.sg_object.get_all_gene_clusters():
                if target_gene_cluster != self.input_bgc:
                    pg.update(
                        task,
                        description=(
                            f"\nComparing {self.input_bgc_id} to {target_gene_cluster.parent.external_id}:{min([i.start for i in target_gene_cluster.features])}-{max([i.start for i in target_gene_cluster.features])}"
                        ),
                    )
                    link_df = pd.concat(
                        [
                            link_df,
                            self._create_link_df(
                                query_gene_cluster=self.input_bgc,
                                target_gene_cluster=target_gene_cluster,
                                tool=tool,
                                **kwargs,
                            ),
                        ]
                    )

                pg.update(task, advance=1, description="[blue]Complete Task")
        self.link_df = link_df
        log.info(f"Finish: Creating links; {len(link_df)} links produced")
        if self.link_df.empty:
            log.warning("No links produced")

    def _choose_group(
        self,
    ):
        log.info("Start: Assigning target BGC proteins to input BGC groups")
        if self.link_df.empty:
            log.info("Finish: Assigning target BGC proteins to input BGC groups")
            log.warning("No links to group by, no groups produced")
            return None
        self.group_df = (
            self.link_df.sort_values(by="score", ascending=False)
            .groupby(["query"], observed=False)
            .head(1)
        )
        log.info("Finish: Assigning target BGC proteins to input BGC groups")

    def add_info_to_cluster_names_for_clustermapjs(self):
        # Modify names of assemblies for their appearance in clustermap
        for k, v in self.sg_object.assemblies.items():
            if v.taxonomy.genus_:
                v.name = f"{v.uid}_{v.taxonomy.genus_}_{v.metadata.all_attributes()['culture_collection']}"
            else:
                v.name = f"{v.uid}_{v.metadata.all_attributes()['culture_collection']}"
            try:
                n_distinct_input_proteins = len(self.input_assembly.feature_uid_set)
                unique_queries_per_assembly = (
                    self.bgc_df[self.bgc_df["assembly_uid"] == k]
                    .groupby("assembly_uid")["query"]
                    .nunique()
                    .values[0]
                )
                a_perc = round(
                    unique_queries_per_assembly / len(n_distinct_input_proteins) * 100,
                    0,
                )
                n_hits = (
                    self.bgc_dfbgc_indices[self.bgc_dfbgc_indices["assembly_uid"] == k]
                    .filter(["cluster", "cluster_unique_hits"])
                    .drop_duplicates()["cluster_unique_hits"]
                    .sum()
                )
                n_perc = round(n_hits / len(n_distinct_input_proteins) * 100, 0)
                v.name = f"{v.name}_{str(n_perc)}%_{str(a_perc)}%"
            except Exception as e:
                log.debug(e)
                pass

    def _count_unique_hits_per(self, column):
        log.info(f"Counting unique hits per {column}")
        return (
            self.working_search_results_df.groupby(column)["query"]
            .nunique()
            .reset_index()
        )

    def _filter_assemblies(self, threshold: float):
        column = "assembly_uid"
        self._filter(column=column, threshold=threshold)

    def _filter_nucleotides(self, threshold: float):
        column = "nucleotide_uid"
        self._filter(column=column, threshold=threshold)

    def _filter(self, column, threshold: float):
        log.info(f"Filtering on {column} where unique hits >= {threshold}")
        if threshold > 0 and threshold < 1:
            threshold = ceil(self.n_searched_proteins * threshold)
        counts = self._count_unique_hits_per(column=column)
        self.working_search_results_df = self.working_search_results_df.merge(
            counts[counts["query"] >= threshold][column],
            how="inner",
            on=column,
        )
        log.info(
            f"{self.working_search_results_df['assembly_uid'].nunique()} assemblies, {self.working_search_results_df['nucleotide_uid'].nunique()} nucleotide sequences had {column}s with >= {threshold} unique query hits"
        )

    def _count_unique_hits_per_cluster(self, df):
        return (
            df.groupby(["nucleotide_uid", "cluster"])["query"]
            .nunique()
            .sort_values(ascending=False)
        )

    def _sort_genes_by_start(self):
        log.info("Sorting genes by start position")
        self.working_search_results_df = self.working_search_results_df.sort_values(
            by=["n_start"], ascending=[True], na_position="first"
        ).reset_index(inplace=False, drop=True)

    def label_clusters(
        self,
    ):
        """Walks through the df and for each nucleotide sequence, assigns genes/protein to a cluster, breaking on gap tolerance

        Args:
            break_bgc_on_gap_of (int, optional): Breaks a "cluster" when the diff of two proteins' starts is greater than this. Defaults to 20000.

        Returns:
            pd.DataFrame: input df + cluster column (integer)
        """
        log.info(
            f"Grouping protein hits if less than {str(self.break_bgc_on_gap_of)} bp apart"
        )
        self._sort_genes_by_start()
        # have to check if groupby output is a df or series and adjust accordingly
        # https://stackoverflow.com/questions/37715246/pandas-groupby-apply-behavior-returning-a-series-inconsistent-output-type
        n_groups = self.working_search_results_df["nucleotide_uid"].nunique()
        if n_groups == 1:
            self.working_search_results_df["cluster"] = (
                self.working_search_results_df["n_start"]
                - self.working_search_results_df["n_end"].shift().fillna(0).astype(int)
                > self.break_bgc_on_gap_of
            ).cumsum()
        elif n_groups > 1:
            self.working_search_results_df["cluster"] = (
                self.working_search_results_df.groupby("nucleotide_uid")
                .apply(
                    lambda group: group["n_start"]
                    > (group["n_end"].shift() + self.break_bgc_on_gap_of)
                )
                .cumsum()
                .reset_index(level=[0])[0]
            )
        else:
            raise ValueError("Huh, expected one or more groups")

    def _collapse_cluster(self, df):
        """Collapse a cluster to a single row

        Args:
            df (pd.DataFrame): input df

        Returns:
            pd.DataFrame: collapsed df
        """
        return (
            df.groupby(["assembly_uid", "nucleotide_uid", "cluster"])
            .agg({"n_start": "min", "n_end": "max"})
            .reset_index()
        )

    def _filter_clusters(self, df, threshold) -> pd.Index:
        """Filter the results for potential BGCs
        Args:
        Returns:
            pd.Index: indices of potential BGCs
        """
        if threshold > 0 and threshold < 1:
            threshold = ceil(self.n_searched_proteins * threshold)

        tempdf = self._count_unique_hits_per_cluster(df)

        tempdf = tempdf[tempdf >= threshold]

        # filter the df so that only rows with the same nucleotide_uid and cluster as the tempdf are kept
        df = df[df.set_index(["nucleotide_uid", "cluster"]).index.isin(tempdf.index)]
        return df

    def write_clustermap_json(self, outpath):
        cmap = SerializeToClustermap(
            sorted_bgcs=self.sorted_bgcs, sg_object=self.sg_object, link_df=self.link_df
        )
        cmap.write(outpath=outpath)

    def user_friendly_hit_df(self, truncate_description_to_n_chars=20):
        with GraphDriver() as db:
            db_res = db.run(
                """
               WITH $input as inputs
               unwind inputs as input
               MATCH (n1:nucleotide {uid:input.nucleotide_uid})-[e1:ENCODES {start:input.n_start}]->(p1:protein {uid:input.target})
               RETURN input.assembly_uid as t_assembly_uid,
                      n1.external_id as t_external_id,
                      e1.locus_tag as t_locus_tag,
                      e1.protein_id as t_protein_id,
                      e1.start as t_start,
                      e1.end as t_end,
                      e1.description as t_description,
                      input.query as query,
                      input.cluster as cluster
             """,
                input=self.working_search_results_df.to_dict("records"),
            ).to_df()
        query_df = pd.DataFrame(
            [
                {
                    "query": i.uid,
                    "q_description": i.description,
                    "q_external_id": i.external_id,
                }
                for i in self.input_bgc.features
            ]
        )
        db_res = pd.merge(db_res, query_df, on="query", how="left")
        db_res = db_res.drop(["query"], axis=1)
        db_res = db_res[
            [
                "t_assembly_uid",
                "t_external_id",
                "t_locus_tag",
                "cluster",
                "q_external_id",
                "t_protein_id",
                "t_start",
                "t_end",
                "t_description",
                "q_description",
            ]
        ]
        db_res = db_res.sort_values(["t_assembly_uid", "t_external_id", "t_start"])
        db_res["t_description"] = db_res["t_description"].apply(
            lambda x: truncate_string(x, truncate_description_to_n_chars)
        )
        db_res["q_description"] = db_res["q_description"].apply(
            lambda x: truncate_string(x, truncate_description_to_n_chars)
        )
        return db_res

    def rich_table(
        self,
        df,
        title="",
        n=100,
        justify="center",
        style="cyan",
        no_wrap=True,
        **kwargs,
    ):
        table = Table(title=title)
        for i in df.columns:
            table.add_column(
                i, justify=justify, style=style, no_wrap=no_wrap, ratio=1, **kwargs
            )
        ii = 0
        for i in df.values:
            table.add_row(*[str(i) for i in i])
            ii += 1
            if ii == n:
                break
        console = CONSOLE
        console.print(table)



"""https://gist.github.com/OsKaR31415/955b166f4a286ed427f667cb21d57bfd"""


def markdown_table_from_list(array, align: str = None):
    """
    Args:
        array: The array to make into a table. Mush be a rectangular array
               (constant width and height).
        align: The alignment of the cells : 'left', 'center' or 'right'.
    """
    # make sure every elements are strings
    array = [[str(elt) for elt in line] for line in array]
    # get the width of each column
    widths = [max(len(line[i]) for line in array) for i in range(len(array[0]))]
    # make every width at least 3 colmuns, because the separator needs it
    widths = [max(w, 3) for w in widths]
    # center text according to the widths
    array = [[elt.center(w) for elt, w in zip(line, widths)] for line in array]

    # separate the header and the body
    array_head, *array_body = array

    header = "| " + " | ".join(array_head) + " |"

    # alignment of the cells
    align = str(align).lower()  # make sure `align` is a lowercase string
    if align == "none":
        # we are just setting the position of the : in the table.
        # here there are none
        border_left = "| "
        border_center = " | "
        border_right = " |"
    elif align == "center":
        border_left = "|:"
        border_center = ":|:"
        border_right = ":|"
    elif align == "left":
        border_left = "|:"
        border_center = " |:"
        border_right = " |"
    elif align == "right":
        border_left = "| "
        border_center = ":| "
        border_right = ":|"
    else:
        raise ValueError("align must be 'left', 'right' or 'center'.")
    separator = (
        border_left + border_center.join(["-" * w for w in widths]) + border_right
    )

    # body of the table
    body = [""] * len(array)  # empty string list that we fill after
    for idx, line in enumerate(array[1:]):
        # for each line, change the body at the correct index
        body[idx] = "| " + " | ".join(line) + " |"
    body = "\n".join(body)

    return header + "\n" + separator + "\n" + body


# T = [["hi", "hello", "How do you do ?"],
#      [1, 1, 1],
#      [2, 12345678, 2],
#      ['contents', 42, 73]]

# print(make_markdown_table(T, align='left'))



import re
import xml.etree.ElementTree as ET
from pathlib import Path
from uuid import uuid4

import numpy as np
import pandas as pd

from socialgene.addons.gnps_library.nr import (
    GnpsLibrarySpectrumNode,
    GnpsLibrarySpectrumToNPClassifierClass,
    GnpsLibrarySpectrumToNPClassifierPathway,
    GnpsLibrarySpectrumToNPClassifierSuperclass,
    GnpsLibraryToChem,
)
from socialgene.addons.gnps_networking.nr import (
    ClusterNode,
    ClusterToSpectrum,
    LibraryHitRel,
    MassSpecFileNode,
    MassSpecFileToAssembly,
    MassSpecFileToSpectrum,
    MolecularNetwork,
    SpectrumNode,
)
from socialgene.addons.npclassifier.nr import (
    NPClassifierClass,
    NPClassifierPathway,
    NPClassifierSuperclass,
)
from socialgene.base.chem import ChemicalCompound
from socialgene.neo4j.neo4j import GraphDriver
from socialgene.nextflow.nodes import ASSEMBLY
from socialgene.utils.file_handling import open_read
from socialgene.utils.logging import log


def capture_assembly_id(s, regex):
    match regex:
        case "_":
            pattern = r"^[^_]*"
        case "__":
            pattern = r"^([^_]+)__"
        case _:
            pattern = regex
    match = re.search(pattern, s)
    if match:
        return match.group(0)
    else:
        return s


class GNPS_SNETS:
    """Parses GNPS molecular network results and adds them to a Neo4j database"""

    def __init__(
        self,
        gnps_dirpath: str = None,
        params_xml_path: str = None,
        specnets_path: str = None,
        selfloop_path: str = None,
        clustersummary_path: str = None,
        clusterinfo_path: str = None,
        map_path: str = None,
    ):
        self.gnps_dirpath = gnps_dirpath
        if self.gnps_dirpath:
            self.gnps_dirpath = Path(self.gnps_dirpath)
        self.params_xml_path = params_xml_path
        self.specnets_path = specnets_path
        self.selfloop_path = selfloop_path
        self.clustersummary_path = clustersummary_path
        self.clusterinfo_path = clusterinfo_path
        self.map_path = map_path
        self.gnps_file_mapping = {}
        self.clusterinfo_df = pd.DataFrame()
        self.specnets_df = pd.DataFrame()
        self.selfloop_df = pd.DataFrame()
        self.clustersummary_df = pd.DataFrame()
        self.workflow_uuid = str(uuid4()) + "_sg_assigned"
        self.task = str(uuid4()) + "_sg_assigned"
        self.library_nodes = set()
        for i in [
            "specnets_path",
            "selfloop_path",
            "clustersummary_path",
            "clusterinfo_path",
            "map_path",
        ]:
            if getattr(self, i):
                setattr(self, i, Path(getattr(self, i)))

    def _parse(self):
        self._get_gnps_filemapping()
        self._workflow_uuid()
        self._task_id()
        self._parse_clusterinfo()
        self._parse_dfs()

    def _find_gnps_paths(self):
        """Parses the downloaded GNPS results for the necessary files, must be unzipped"""
        params_xml_path = Path(self.gnps_dirpath, "params.xml")
        specnets_path = list(self.gnps_dirpath.glob("**/result_specnets_DB/*.tsv"))
        clustersummary_path = list(
            self.gnps_dirpath.glob(
                "**/clusterinfosummarygroup_attributes_withIDs_withcomponentID/*.clustersummary"
            )
        )
        selfloop_path = list(self.gnps_dirpath.glob("**/networkedges_selfloop/*"))
        clusterinfo_path = list(self.gnps_dirpath.glob("**/clusterinfo/*.clusterinfo"))
        for i in [specnets_path, clustersummary_path, selfloop_path, clusterinfo_path]:
            if len(i) > 1:
                raise ValueError(f"More than one file found for {i}")
            if len(i) == 0:
                raise ValueError(f"No file found for {i}")
        specnets_path = specnets_path[0]
        clustersummary_path = clustersummary_path[0]
        selfloop_path = selfloop_path[0]
        clusterinfo_path = clusterinfo_path[0]
        for i in [
            specnets_path,
            clustersummary_path,
            selfloop_path,
            params_xml_path,
            clusterinfo_path,
        ]:
            if not i.exists():
                raise ValueError(f"File not found for {i}")
        self.params_xml_path = params_xml_path
        self.specnets_path = specnets_path
        self.selfloop_path = selfloop_path
        self.clustersummary_path = clustersummary_path
        self.clusterinfo_path = clusterinfo_path

    def _get_gnps_filemapping(self):
        with open_read(self.params_xml_path) as f:
            xml_root = ET.fromstring(f.read())
        for i in xml_root.iter("parameter"):
            if i.attrib["name"] == "upload_file_mapping":
                try:
                    self.gnps_file_mapping[i.text.split("|")[0]] = i.text.split("/")[-1]
                except Exception as e:
                    log.debug(f"Error parsing file mapping: {e}")

    def _parse_clusterinfo(self):
        """Parses the clusterinfo file and replaces the mapped filename with the original filename"""
        with open_read(self.clusterinfo_path) as f:
            clusterinfo_df = pd.read_csv(f, sep="\t").replace(np.nan, None)
        clusterinfo_df.columns = [
            name.lower()
            .replace("#", "")
            .replace(" ", "_")
            .replace("(", "")
            .replace(")", "")
            .replace("/", "_")
            .replace("-", "_")
            for name in clusterinfo_df.columns
        ]
        # replaces the mapped filename with the original filename
        clusterinfo_df["original_filename"] = clusterinfo_df["filename"].apply(
            lambda x: self.gnps_file_mapping[x.split("/")[-1]]
        )
        self.clusterinfo_df = clusterinfo_df

    def _workflow_uuid(self):
        """Parses the workflow uuid from the params.xml file"""
        with open_read(self.params_xml_path) as f:
            for i in f:
                if 'parameter name="uuid"' in i:
                    self.workflow_uuid = i.split(">")[1].split("<")[0]

    def _task_id(self):
        """Parses the task id from the params.xml file"""
        with open_read(self.params_xml_path) as f:
            for i in f:
                if 'parameter name="task"' in i:
                    self.task = i.split(">")[1].split("<")[0]

    def _parse_dfs(self):
        """Parses the specnets, selfloop, and clustersummary files into pandas dataframes and sanitizes the column names with lowercase and underscores"""
        for i in ["specnets_path", "selfloop_path", "clustersummary_path"]:
            with open_read(getattr(self, i)) as f:
                df = pd.read_csv(f, sep="\t").replace(np.nan, None)
            df.columns = [
                name.lower()
                .replace("#", "")
                .replace(" ", "_")
                .replace("(", "")
                .replace(")", "")
                .replace("/", "_")
                .replace("-", "_")
                for name in df.columns
            ]
            setattr(self, f"{i.split('_')[0]}_df", df)
        try:
            self.clustersummary_df.drop(columns=["uniquefilesources"], inplace=True)
        except KeyError:
            log.debug("No column 'uniquefilesources' in clustersummary file")
        try:
            self.specnets_df.drop(
                columns=[
                    "updateworkflowname",
                    "scan",
                    "spectrumfile",
                    "libraryname",
                    "filescanuniqueid",
                    "numberhits",
                    "tags",
                ],
                inplace=True,
            )
        except KeyError:
            log.debug(
                f"{[i for i in ['updateworkflowname', 'scan', 'spectrumfile', 'libraryname', 'filescanuniqueid', 'numberhits', 'tags'] if i not in self.specnets_df.columns]}"
            )

    @staticmethod
    def _check_db_for_assemblies(assemblies):
        """Checks the Neo4j database for assemblies that can be linked to the GNPS results"""
        with GraphDriver() as db:
            results = db.run(
                """
                WITH $assemblies as assemblies
                UNWIND assemblies as assembly
                MATCH (a1:assembly {uid: assembly})
                RETURN assembly
                """,
                assemblies=list(assemblies),
            ).value()
        if results:
            log.info(
                f"Assemblies in GNPS results found in db: {len(results)} of {len(assemblies)}"
            )
            log.info(
                f"Assemblies in GNPS results not found in db: {set(assemblies) - set(results)}"
            )
        else:
            log.warning(
                "No assemblies in GNPS results matched to assemblies in the database"
            )
        return results

    # Add nodes to Neo4j
    def _add_input_spectra(self, create=False):
        def temp(i):
            spec = SpectrumNode()
            d = i | {"workflow_uuid": self.workflow_uuid}
            spec.fill_from_dict(d)
            return spec

        SpectrumNode.add_multiple_to_neo4j(
            [temp(i) for i in self.clusterinfo_df.to_dict("records")], create=create
        )

    def _add_input_ms_files(self, create=False):
        def temp(i):
            spec = MassSpecFileNode()
            d = {
                "workflow_uuid": self.workflow_uuid,
                "gnps_filename": i["filename"],
                "filename": i["original_filename"],
            }
            spec.fill_from_dict(d)
            return spec

        MassSpecFileNode.add_multiple_to_neo4j(
            [temp(i) for i in self.clusterinfo_df.to_dict("records")], create=create
        )

    def _create_library_hit_nodes(self):
        """Creates GNPS library hit nodes"""
        for i in self.specnets_df.to_dict("records"):
            try:
                a = GnpsLibrarySpectrumNode()
                a.fill_from_dict(i)
                a.properties["uid"] = str(i["spectrumid"])
                self.library_nodes.add(a)
            except Exception as e:
                log.debug(f"Error creating library hit node for {i['spectrumid']}: {e}")

    def _add_library_hit_nodes(self, create=False):
        """Adds GNPS library hit nodes to the Neo4j database"""
        list(self.library_nodes)[0].add_multiple_to_neo4j(
            list(self.library_nodes), create=create
        )

    def _add_cluster_nodes(self, create=False):
        clusternodes = set()
        for i in self.clustersummary_df.to_dict("records"):
            a = ClusterNode()
            d = i | {"workflow_uuid": self.workflow_uuid, "task": self.task}
            a.fill_from_dict(d)
            clusternodes.add(a)
        clusternodes = list(clusternodes)
        clusternodes[0].add_multiple_to_neo4j(clusternodes, create=create)

    # Add rels to Neo4j

    def _link_cluster_to_spectrum(self):
        rels = set()
        for i in self.clusterinfo_df.to_dict("records"):
            cn = ClusterNode()
            cn.fill_from_dict(
                {
                    "cluster_index": i["clusteridx"],
                    "workflow_uuid": self.workflow_uuid,
                    "task": self.task,
                }
            )
            sn = SpectrumNode()
            sn.fill_from_dict(i | {"workflow_uuid": self.workflow_uuid})
            rels.add(ClusterToSpectrum(start=sn, end=cn))
        rels = list(rels)
        rels[0].add_multiple_to_neo4j(rels, create=False)

    def _validate_map(self):
        # columns: assembly_id, mass_spec_filename
        if not self.map_path or not Path(self.map_path).exists():
            raise ValueError("No map file found")
        with open_read(self.map_path) as f:
            map_df = pd.read_csv(
                f,
                sep="\t",
                header=None,
            ).replace(np.nan, None)
        # should only have two columns
        if len(map_df.columns) != 2:
            raise ValueError("Map file should only have two columns")
        map_df.columns = ["assembly", "mass_spec_file"]
        # get first column as list
        assemblies = set(map_df.iloc[:, 0])
        matched_assemblies = self._check_db_for_assemblies(assemblies)
        # filter df for matched assemblies
        map_df = map_df[map_df.iloc[:, 0].isin(matched_assemblies)]
        # change second column to Path().name
        map_df.iloc[:, 1] = map_df.iloc[:, 1].apply(lambda x: Path(x).name)
        return map_df

    def _link_ms_file_to_assembly(self):
        map_df = self._validate_map()
        # merge map_df and clusterinfo_df, merge on mass_spec_file and original_filename
        df = self.clusterinfo_df.merge(
            map_df, left_on="original_filename", right_on="mass_spec_file", how="inner"
        )
        rels = set()
        for i in df.to_dict("records"):
            cn = MassSpecFileNode()
            d = {
                "workflow_uuid": self.workflow_uuid,
                "gnps_filename": i["filename"],
                "filename": i["original_filename"],
            }
            cn.fill_from_dict(d)
            sn = ASSEMBLY()
            d = {"uid": i["assembly"]}
            sn.fill_from_dict(d)
            rels.add(MassSpecFileToAssembly(start=cn, end=sn))
        rels = list(rels)
        rels[0].add_multiple_to_neo4j(rels, create=False)

    def _link_spectrum_to_ms_file(self):
        rels = set()
        for i in self.clusterinfo_df.to_dict("records"):
            cn = MassSpecFileNode()
            d = {
                "workflow_uuid": self.workflow_uuid,
                "gnps_filename": i["filename"],
                "filename": i["original_filename"],
            }
            cn.fill_from_dict(d)
            sn = SpectrumNode()
            d = i | {"workflow_uuid": self.workflow_uuid}
            sn.fill_from_dict(d)
            rels.add(MassSpecFileToSpectrum(start=cn, end=sn))
        rels = list(rels)
        rels[0].add_multiple_to_neo4j(rels, create=False)

    def _link_npclassifiers(self):
        """Classifies GNPS library hits by linking them to NPClassifier nodes"""
        to_loop = [
            (
                "npclassifier_superclass",
                NPClassifierSuperclass,
                GnpsLibrarySpectrumToNPClassifierSuperclass,
            ),
            (
                "npclassifier_class",
                NPClassifierClass,
                GnpsLibrarySpectrumToNPClassifierClass,
            ),
            (
                "npclassifier_pathway",
                NPClassifierPathway,
                GnpsLibrarySpectrumToNPClassifierPathway,
            ),
        ]
        for name, nodeclass, relclass in to_loop:
            if name not in self.specnets_df.columns:
                # not all GNPS results will have npclassifier results
                continue
            df = self.specnets_df[~self.specnets_df[name].isnull()]
            df = df[["spectrumid", name]]
            npc_set = set()
            gnp_set = set()
            rel_set = set()
            for i in df.to_dict("records"):
                npc = nodeclass(properties={"uid": i[name]})
                gnp = GnpsLibrarySpectrumNode(properties={"uid": i["spectrumid"]})
                rel = relclass(start=gnp, end=npc)
                # accumulate unique nodes and relationships to bulk add
                npc_set.add(npc)
                gnp_set.add(gnp)
                rel_set.add(rel)
            npc_set = list(npc_set)
            gnp_set = list(gnp_set)
            rel_set = list(rel_set)
            for i in [npc_set, gnp_set, rel_set]:
                i[0].add_multiple_to_neo4j(i, create=False)

    def _link_cluster_to_library_no_spec_id(self):
        if "spectrumid" not in self.clustersummary_df.columns:
            if "libraryid" not in self.clustersummary_df.columns:
                log.warn("No libraryid or spectrumid in clustersummary file")
                return
        df = self.clustersummary_df[["libraryid", "cluster_index"]]
        df = df[~df["libraryid"].isnull()]
        s_set = set()
        e_set = set()
        r_set = set()
        for i in df.to_dict("records"):
            s = ClusterNode(
                properties={
                    "cluster_index": i["cluster_index"],
                    "workflow_uuid": self.workflow_uuid,
                    "task": self.task,
                }
            )
            e = GnpsLibrarySpectrumNode(properties={"uid": i["libraryid"]})
            r = LibraryHitRel(start=s, end=e)
            s_set.add(s)
            e_set.add(e)
            r_set.add(r)
        s_set = list(s_set)
        e_set = list(e_set)
        r_set = list(r_set)
        for i in [s_set, e_set, r_set]:
            i[0].add_multiple_to_neo4j(i, create=False)

    def _link_cluster_to_library(self):
        if "spectrumid" not in self.clustersummary_df.columns:
            # not all GNPS results will have library hits
            log.warn(
                "Older versions of GNPS results may have library ids but not spectrumid? Skipping."
            )
            try:
                self._link_cluster_to_library_no_spec_id()
            except Exception as e:
                raise ValueError(f"Failed to link clusters to library hits: {e}")
            return
        df = self.clustersummary_df[["spectrumid", "cluster_index"]]
        df = df[~df["spectrumid"].isnull()]
        s_set = set()
        e_set = set()
        r_set = set()
        for i in df.to_dict("records"):
            s = ClusterNode(
                properties={
                    "cluster_index": i["cluster_index"],
                    "workflow_uuid": self.workflow_uuid,
                    "task": self.task,
                }
            )
            e = GnpsLibrarySpectrumNode(properties={"uid": i["spectrumid"]})
            r = LibraryHitRel(start=s, end=e)
            s_set.add(s)
            e_set.add(e)
            r_set.add(r)
        s_set = list(s_set)
        e_set = list(e_set)
        r_set = list(r_set)
        for i in [s_set, e_set, r_set]:
            i[0].add_multiple_to_neo4j(i, create=False)

    def _link_network(self, create=False):
        s_set = set()
        e_set = set()
        r_set = set()
        for i in self.selfloop_df.to_dict("records"):
            s = ClusterNode(
                properties={
                    "cluster_index": i["clusterid1"],
                    "workflow_uuid": self.workflow_uuid,
                    "task": self.task,
                }
            )
            e = ClusterNode(
                properties={
                    "cluster_index": i["clusterid2"],
                    "workflow_uuid": self.workflow_uuid,
                    "task": self.task,
                }
            )
            r = MolecularNetwork(start=s, end=e)

            if "edgeannotation" not in i or i["edgeannotation"] == " ":
                i["edgeannotation"] = None
            r.fill_from_dict(i)
            s_set.add(s)
            e_set.add(e)
            r_set.add(r)
        s_set = list(s_set)
        e_set = list(e_set)
        r_set = list(r_set)
        for i in [s_set, e_set]:
            # create=False because nodes should already exist so we want to merge not create
            i[0].add_multiple_to_neo4j(i, create=False)
        r_set[0].add_multiple_to_neo4j(r_set, create=create)

    def _link_library_to_chem(self):
        all_chem_nodes = set()
        all_rels = set()
        for i in self.library_nodes:
            chemstring = None
            if "inchi" in i.properties and i.properties["inchi"].startswith("InChI"):
                chemstring = i.properties["inchi"]
            elif "smiles" in i.properties and i.properties["smiles"]:
                chemstring = i.properties["smiles"]
            if chemstring:
                try:
                    cmpd = ChemicalCompound(chemstring)
                    a = cmpd.node
                    all_chem_nodes.add(a)
                    all_rels.add(GnpsLibraryToChem(start=i, end=a))
                except Exception as e:
                    log.debug(
                        f"Error creating chemical compound node for {i.properties['uid']}: {e}"
                    )
        all_chem_nodes = list(all_chem_nodes)
        all_rels = list(all_rels)
        if all_chem_nodes:
            all_chem_nodes[0].add_multiple_to_neo4j(all_chem_nodes, create=False)
        if all_rels:
            all_rels[0].add_multiple_to_neo4j(all_rels, create=False)



import csv
from pathlib import Path
from typing import List

from socialgene.neo4j.schema.modules import Modules
from socialgene.utils.logging import log


class SocialgeneModules(Modules):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.selected_nodes = set()
        self.selected_relationships = set()

    def add_modules(self, module_list: List[str]):
        """Take a list of modules and create sets of corresponding node/relationship objects

        Args:
            module_list (List[str]): list of socialgene modules (e.g. ["base", "protein"])
        """
        # if only a single module str is provided, make sure it's a list
        for module in list(module_list):
            self.add_module(module)

    def add_module(self, module_id: str):
        if module_id not in self.modules:
            raise ValueError(
                f"Module {module_id} not found. Please select from: {list(self.modules.keys())}"
            )
        else:
            log.debug(f"Adding module: {module_id}")
            for node in self.modules.get(module_id).nodes:
                self.selected_nodes.add(node)
            for rel in self.modules.get(module_id).relationships:
                self.selected_relationships.add(rel)

    def _writer(self, outdir, header, header_filename):
        outpath = Path(outdir, header_filename)
        with open(outpath, "w") as tsv_output_con:
            tsv_writer = csv.writer(
                tsv_output_con,
                delimiter="\t",
                quotechar='"',
                quoting=csv.QUOTE_MINIMAL,
            )
            tsv_writer.writerow(header)
            log.info(f"\tWriting {header_filename} to: {outpath}")

    def write_neo4j_headers(self, outdir: str):
        for node in self.selected_nodes:
            self._writer(
                outdir,
                header=node().header,
                header_filename=node().header_filename,
            )
        for rel in self.selected_relationships:
            self._writer(
                outdir,
                header=rel.header,
                header_filename=rel.header_filename,
            )



from socialgene.addons.gnps_library.nr import GnpsLibrarySpectrumNode
from socialgene.neo4j.neo4j_element import Node, Relationship
from socialgene.nextflow.nodes import ASSEMBLY


class ClusterNode(Node):
    neo4j_label = ["gnps_cluster"]
    description = "Represents a GNPS molecular networking cluster"
    required_properties = ["cluster_index", "workflow_uuid", "task"]
    property_specification = {
        "task": str,
        "defaultgroups": str,
        "g1": int,
        "g2": int,
        "g3": int,
        "g4": int,
        "g5": int,
        "g6": int,
        "gnpslinkout_cluster": str,
        "gnpslinkout_network": str,
        "mqscore": float,
        "mzerrorppm": float,
        "massdiff": float,
        "rtmean": float,
        "rtmean_min": float,
        "rtstderr": float,
        "uniquefilesources": str,
        "uniquefilesourcescount": int,
        "cluster_index": int,
        "componentindex": int,
        "number_of_spectra": int,
        "parent_mass": float,
        "precursor_charge": int,
        "precursor_mass": float,
        "sumprecursor_intensity": float,
        "workflow_uuid": str,
    }
    constraints_unique = ["cluster_index", "workflow_uuid", "task"]


class SpectrumNode(Node):
    neo4j_label = ["ms2_spectrum"]
    description = "Represents a GNPS molecular networking spectrum"
    property_specification = {
        "specidx": str,
        "original_filename": str,
        "parentmass": float,
        "charge": int,
        "rettime": float,
        "workflow_uuid": str,
    }
    constraints_unique = ["original_filename", "specidx", "workflow_uuid"]


class MassSpecFileNode(Node):
    neo4j_label = ["mass_spectrum_file"]
    description = "Represents a GNPS molecular networking spectrum file"
    property_specification = {
        "filename": str,
        "gnps_filename": str,
        "workflow_uuid": str,
    }
    constraints_unique = ["original_filename", "workflow_uuid"]


class MassSpecFileToSpectrum(Relationship):
    neo4j_label = "HAS"
    description = "Connects a GNPS mass spectrum file to a GNPS spectrum"
    start_class = MassSpecFileNode
    end_class = SpectrumNode
    property_specification = {}
    required_properties = []


class MassSpecFileToAssembly(Relationship):
    neo4j_label = "ANALYSIS_OF"
    description = "Connects a GNPS mass spectrum file to an assembly"
    start_class = MassSpecFileNode
    end_class = ASSEMBLY
    property_specification = {}
    required_properties = []


class ClusterToSpectrum(Relationship):
    neo4j_label = "CLUSTERS_TO"
    description = "Connects a GNPS cluster to a GNPS spectrum"
    start_class = SpectrumNode
    end_class = ClusterNode
    property_specification = {}
    required_properties = []


class LibraryHitRel(Relationship):
    neo4j_label = "LIBRARY_HIT"
    description = "Connects a GNPS cluster to a GNPS library hit"
    start_class = ClusterNode
    end_class = GnpsLibrarySpectrumNode


class MolecularNetwork(Relationship):
    neo4j_label = "MOLECULAR_NETWORK"
    description = "Connects the GNPS molecular networks"
    start_class = ClusterNode
    end_class = ClusterNode
    property_specification = {
        "delta_mz": float,
        "meh": float,
        "cosine": float,
        "otherscore": float,
        "componentindex": int,
        "edgeannotation": str,
    }
    required_properties = []



from typing import List

from socialgene.clustermap.serialize import SerializeToClustermap
from socialgene.search.hmmer import SearchDomains
from socialgene.utils.logging import log


def search_bgc(
    input,
    hmm_dir: str = None,
    use_neo4j_precalc: bool = True,
    assemblies_must_have_x_matches: float = 0.6,
    nucleotide_sequences_must_have_x_matches: float = 0.6,
    gene_clusters_must_have_x_matches: float = 0.6,
    break_bgc_on_gap_of: int = 10000,
    target_bgc_padding: int = 20000,
    max_domains_per_protein: float = 3,
    max_outdegree: int = 100000,
    max_query_proteins: int = 10,
    scatter: bool = False,
    locus_tag_bypass_list: List[str] = None,
    protein_id_bypass_list: List[str] = None,
    only_culture_collection: bool = False,
    frac: float = 0.75,
    run_async: bool = True,
    analyze_with: str = "blastp",
    outpath_clinker: str = None,
    blast_speed: str = "fast",
    blast_threads: int = 1,
    adaptive_padding: bool = True,
):
    log.info(f"Running search with args: {locals()}")

    match blast_speed:
        case "fast":
            blastarg = "--fast"
        case "sensitive":
            blastarg = "--sensitive"
        case "ultra-sensitive":
            blastarg = "--ultra-sensitive"
        case _:
            raise ValueError(f"Invalid blast speed: {blast_speed}")

    search_object = SearchDomains(
        input=input,
        hmm_dir=hmm_dir,
        use_neo4j_precalc=use_neo4j_precalc,
        assemblies_must_have_x_matches=assemblies_must_have_x_matches,
        nucleotide_sequences_must_have_x_matches=nucleotide_sequences_must_have_x_matches,
        gene_clusters_must_have_x_matches=gene_clusters_must_have_x_matches,
        break_bgc_on_gap_of=break_bgc_on_gap_of,
        target_bgc_padding=target_bgc_padding,
        adaptive_padding=adaptive_padding,
    )
    search_object.outdegree_table
    search_object.prioritize_input_proteins(
        max_domains_per_protein=max_domains_per_protein,
        max_outdegree=max_outdegree,
        max_query_proteins=max_query_proteins,
        scatter=scatter,
        locus_tag_bypass_list=locus_tag_bypass_list,
        protein_id_bypass_list=protein_id_bypass_list,
    )
    search_object.outdegree_table
    search_object.search(
        only_culture_collection=only_culture_collection, frac=frac, run_async=run_async
    )
    # filters assemblies and nucleotide seqs that fall below threshold % of query proteins
    if search_object.raw_search_results_df.empty:
        log.warning("Stopping, no search results to process")
        raise
    search_object.filter()
    # labels clusters with a unique id based on break_bgc_on_gap_of
    search_object.label_clusters()
    df = search_object._primary_bgc_regions()
    log.info(
        f"First pass resulted in {df.assembly_uid.nunique()} assemblies, {df.nucleotide_uid.nunique()} nucleotide sequences had {df.cluster.nunique()} putative BGCs"
    )

    df["n_start"] = df["n_start"] - search_object.search_padding
    df["n_end"] = df["n_end"] + search_object.search_padding
    search_object._bgc_regions_to_sg_object(df)

    ########
    # add bgcs as gene_clusters to the locus objects
    for i, row in df.iterrows():
        search_object.sg_object.assemblies[row["assembly_uid"]].get_locus_by_uid(
            row["nucleotide_uid"]
        ).add_bgcs_by_start_end(
            start=row["n_start"] - search_object.search_padding,
            end=row["n_end"] + search_object.search_padding,
            uid=row["cluster"],
        )
    # add input bgc as gene_cluster to the locus objects
    if analyze_with == "hmmer":
        _ = search_object.sg_object.annotate_proteins_with_neo4j(
            protein_uids=None, annotate_all=True, progress=False
        )
    else:
        search_object.sg_object.add_sequences_from_neo4j()

    # Find RBH between input BGC and putative BGCs
    if analyze_with == "blastp":
        argstring = f"{blastarg} --max-hsps 1"
    else:
        argstring = ""
    search_object._create_links(
        tool=analyze_with, argstring=argstring, threads=blast_threads
    )
    # Assign protein groups for the clinker plot legend
    search_object._choose_group()

    if not outpath_clinker:
        return search_object
    # return search_object
    if gene_clusters_must_have_x_matches > 1:
        # scale to 0-1
        gene_clusters_must_have_x_matches = gene_clusters_must_have_x_matches / len(
            search_object.input_bgc.proteins
        )
    for group, df in search_object.link_df.groupby("target_gene_cluster"):
        minstart = (
            df.target_feature.apply(lambda x: x.start).min()
            - search_object.target_bgc_padding
        )
        group.features = {i for i in group.features if i.start > minstart}
        maxend = (
            df.target_feature.apply(lambda x: x.end).max()
            + search_object.target_bgc_padding
        )
        group.features = {i for i in group.features if i.end < maxend}

    assemblies = search_object._rank_order_bgcs(
        threshold=gene_clusters_must_have_x_matches
    )
    assemblies = [i.parent.parent for i in assemblies]
    ids = [i.uid for i in assemblies]
    # log.warning(ids)
    if len(ids) == 1 and ids[0] == search_object.input_bgc_id:
        raise ValueError("Only match was to self")
    if len(ids) == 0:
        raise ValueError("No matches found")

    assemblies = [search_object.input_assembly] + list(assemblies)[:50]
    z = SerializeToClustermap(
        sg_object=search_object.sg_object,
        sorted_bgcs=assemblies,
        link_df=search_object.link_df,
        group_df=search_object.group_df,
    )
    z.write(outpath_clinker)
    return search_object



from dataclasses import dataclass
from typing import List

import socialgene.nextflow.nodes as nodes
import socialgene.nextflow.relationships as rels


@dataclass
class Single_Module:
    """Class for keeping track of an item in inventory."""

    module_id: str
    nodes: List[str]
    relationships: List[str]


class Modules:
    # These are used in Nextflow, here: https://github.com/socialgene/sgnf/blob/main/subworkflows/local/sg_modules.nf
    # They simply group the different nodes/relationships into somewhat sensible "modules"
    def __init__(self, *args, **kwargs):
        self.modules = {}
        self._add_module(
            module_id="base",
            nodes=[nodes.ASSEMBLY, nodes.NUCLEOTIDE, nodes.PROTEIN],
            relationships=[
                rels.ASSEMBLES_TO,
                rels.ENCODES,
            ],
        )
        self._add_module(
            module_id="go",
            nodes=[nodes.GOTERM],
            relationships=[
                rels.PROTEIN_TO_GO,
                rels.GOTERM_RELS,
            ],
        )
        self._add_module(
            module_id="protein",
            nodes=[nodes.PROTEIN],
            relationships=[],
        )
        self._add_module(
            module_id="parameters",
            nodes=[nodes.PARAMETERS],
            relationships=[],
        )
        self._add_module(
            module_id="base_hmm",
            nodes=[
                nodes.HMM,
                nodes.HMM_SOURCE,
            ],
            relationships=[rels.ANNOTATES, rels.SOURCE_DB],
        )
        self._add_module(
            module_id="ncbi_taxonomy",
            nodes=[nodes.TAXID],
            relationships=[rels.TAXON_PARENT, rels.IS_TAXON],
        )
        self._add_module(
            module_id="tigrfam",
            nodes=[
                nodes.GOTERM,
                nodes.TIGRFAM_MAINROLE,
                nodes.TIGRFAM_SUBROLE,
                nodes.TIGRFAM_ROLE,
            ],
            relationships=[
                rels.MAINROLE_ANN,
                rels.ROLE_ANN,
                rels.SUBROLE_ANN,
                rels.GO_ANN,
            ],
        )
        self._add_module(
            module_id="blastp",
            nodes=[],
            relationships=[rels.BLASTP],
        )
        self._add_module(
            module_id="mmseqs",
            nodes=[],
            relationships=[rels.MMSEQS2],
        )

    def _add_module(self, **kwargs):
        _module_id = kwargs.get("module_id")
        if _module_id in self.modules:
            raise ValueError(f"Module ID: {_module_id} already exists")
        self.modules.update({_module_id: Single_Module(**kwargs)})



# flake8: noqa F401
import builtins
from textwrap import wrap

from rich.table import Table

import socialgene.addons.chebi.nr
import socialgene.addons.chembl.nr
import socialgene.addons.chemistry.nr
import socialgene.addons.classyfire.nr
import socialgene.addons.gnps_library.nr
import socialgene.addons.gnps_networking.nr
import socialgene.addons.mibig.nr
import socialgene.addons.npatlas.nr
import socialgene.addons.npclassifier.nr
import socialgene.addons.npmrd.nr
import socialgene.addons.publication.nr
import socialgene.nextflow.nodes
import socialgene.nextflow.relationships
from socialgene.config import env_vars
from socialgene.neo4j.neo4j_element import Node, Relationship
from socialgene.utils.lists_to_markdown import markdown_table_from_list
from socialgene.utils.logging import CONSOLE, log

builtins.print = print


def recursive_get(x, levels=10):
    if levels == 0:
        return set(x)
    else:
        subclasses = [i.__subclasses__() for i in x]
        flattened_subclasses = [
            subclass for sublist in subclasses for subclass in sublist
        ]
        return set(x) | recursive_get(flattened_subclasses, levels - 1)


class GraphSchema:
    ALL_NODES = recursive_get(Node.__subclasses__())
    ALL_RELATIONSHIPS = recursive_get(Relationship.__subclasses__())
    NEXTFLOW_NODES = [
        x for x in ALL_NODES if x.__module__.startswith("socialgene.nextflow")
    ]
    NEXTFLOW_RELATIONSHIPS = [
        x for x in ALL_RELATIONSHIPS if x.__module__.startswith("socialgene.nextflow")
    ]
    ADDON_NODES = [x for x in ALL_NODES if x.__module__.startswith("socialgene.addons")]
    ADDON_RELATIONSHIPS = [
        x for x in ALL_RELATIONSHIPS if x.__module__.startswith("socialgene.addons")
    ]

    def __init__(self):
        pass

    def _nodes_table(self):
        table = Table(title="Nodes", show_lines=True)
        table.add_column("Label", justify="left", style="cyan", no_wrap=True, ratio=1)
        table.add_column("From", justify="left", style="cyan", no_wrap=True, ratio=1)
        table.add_column(
            "Description",
            justify="left",
            style="cyan",
            no_wrap=False,
            ratio=4,
            max_width=50,
        )
        table.add_column("Properties", style="magenta", ratio=1)
        # sort by label which is the key
        for i in sorted(list(self.ALL_NODES), key=lambda x: x.neo4j_label[0]):
            table.add_row(
                ":".join(i.neo4j_label),
                i.__module__,
                i.description,
                "\n".join(wrap(", ".join(i.property_specification))),
            )
        yield table

    def _relationships_table(self):
        table = Table(title="Relationships", show_lines=True)
        table.add_column("Label", justify="left", style="cyan", no_wrap=True, ratio=1)
        table.add_column(
            "Defined in", justify="left", style="cyan", no_wrap=True, ratio=1
        )
        table.add_column("Relationship", style="magenta", ratio=1)
        table.add_column("NF results subdirectory", style="magenta", ratio=1)
        table.add_column("Neo4j header file", style="magenta", ratio=1)
        for i in sorted(list(self.ALL_RELATIONSHIPS), key=lambda x: x.neo4j_label[0]):
            table.add_row(
                i.neo4j_label,
                i.__module__,
                i(i.start_class(), i.end_class()).__str__(),
                i.target_subdirectory,
                i.header_filename,
            )
        yield table

    def _markdown_table_nodes(nodelist):
        cols = [
            (
                "Label",
                "Description",
                "NF results subdirectory",
                "Neo4j header file",
                "Unique on",
                "properties",
            )
        ]
        rows = [
            (
                ":".join(i.neo4j_label),
                i.description,
                i.target_subdirectory,
                i.header_filename,
                ", ".join(i.constraints_unique),
                [k for k in i.property_specification.keys()],
            )
            for i in sorted(list(nodelist), key=lambda x: x.neo4j_label[0])
        ]
        cols.extend(rows)
        print(
            markdown_table_from_list(
                cols,
                align="left",
            )
        )

    def _markdown_table_rels(rellist):
        cols = [
            (
                "Label",
                "Relationship",
                "NF results subdirectory",
                "Neo4j header file",
            )
        ]
        rows = [
            (
                i.neo4j_label,
                i(i.start_class(), i.end_class()).__str__(),
                i.target_subdirectory,
                i.header_filename,
            )
            for i in sorted(list(rellist), key=lambda x: x.neo4j_label[0])
        ]
        cols.extend(rows)

        print(
            markdown_table_from_list(
                cols,
                align="left",
            )
        )


def main():  # pragma: no cover
    import argparse

    parser = argparse.ArgumentParser(description="Print node info")
    parser.add_argument(
        "--markdown",
        help="",
        default=False,
        required=False,
        action=argparse.BooleanOptionalAction,
    )
    parser.add_argument(
        "--nodes",
        help="",
        default=False,
        required=False,
        action=argparse.BooleanOptionalAction,
    )
    parser.add_argument(
        "--rels",
        help="",
        default=False,
        required=False,
        action=argparse.BooleanOptionalAction,
    )
    args = parser.parse_args()
    if args.markdown:
        if args.nodes:
            GraphSchema._markdown_table_nodes(GraphSchema.ALL_NODES)
        if args.rels:
            GraphSchema._markdown_table_rels(GraphSchema.ALL_RELATIONSHIPS)
    else:
        before_width = CONSOLE.width
        CONSOLE.width = 300
        if args.nodes:
            CONSOLE.print(GraphSchema()._nodes_table().__next__())
            CONSOLE.width = before_width
        if args.rels:
            CONSOLE.print(GraphSchema()._relationships_table().__next__())
        CONSOLE.width = before_width


if __name__ == "__main__":
    main()



import urllib.request
from contextlib import closing
from urllib.request import urlopen


def ping(url):
    return urllib.request.urlopen(url).getcode()


def get_size(url, expected_size):
    with closing(urlopen(url)) as response:
        cl = int(response.headers["Content-length"])
        if abs(expected_size - cl) > 1000:
            raise ValueError(
                f"Expected_size was {expected_size}. The response header claims it is {cl}. Believe what you will."
            )
    return cl



from collections.abc import Generator
from typing import List

import numpy as np

from socialgene.utils.logging import log


def chunk_a_list_with_numpy(input_list: List, n_chunks: int) -> Generator:
    """Chunk a list into n-lists

    Args:
        input_list (list): input list to be chunked
        n_chunks (int): integer of number chunks to split list into
    """
    if len(input_list) < n_chunks:
        log.info(
            "chunk_a_list_with_numpy(): n_chunks is < len(input_list), proceeding with len(input_list)"
        )
        n_chunks = len(input_list)
    return (list(i) for i in np.array_split(input_list, n_chunks))



#
# import gzip
# from mimetypes import guess_type
# from functools import partial
# import sqlite3

#

#
# from socialgene.utils.logging import log
# from socialgene.base.socialgene import SocialGene
# from sqlite3 import Error

# # see assumption note in ./socialgene/src/socialgene/utils/protein_sqlite.py


# class NCBIFeatureTable(SocialGene):
#     def __init__(self):
#         super().__init__()
#         self.conn = None

#     def connect(self, db_path):
#         try:
#             self.conn = sqlite3.connect(db_path)
#         except Error as e:
#             log.error(e)

#     def sql_hash_lookup(self, accession):
#         cur = self.conn.cursor()
#         cur.execute("SELECT sha512t24u FROM protmap WHERE accession=?", ([accession]))
#         rows = cur.fetchall()
#         # ensure only one result was returned
#         if not rows:
#             raise ValueError(
#                 f'Couldn\'t find a match for "{accession}" in the sqlite database'
#             )
#         if not len(rows[0]) == 1:
#             raise ValueError(
#                 f'Found multiple hashes for "{accession}" in the sqlite database'
#             )
#         return rows[0][0]

#     def parse(self, input_path):
#         encoding = guess_type(input_path)[1]
#         if encoding == "gzip":
#             _open = partial(gzip.open, mode="rt")
#         else:
#             _open = open
#         with _open(input_path) as f:
#             for line in f:
#                 if line.startswith("#"):
#                     continue
#                 temp = line.split("\t")
#                 self.add_assembly(id=temp[2])
#                 self.assemblies[temp[2]].add_locus(id=temp[6])
#                 uid = self.sql_hash_lookup(temp[10])
#                 self.assemblies[temp[2]].loci[temp[6]].add_feature(
#                     type="protein",
#                     id=uid,
#                     start=temp[7],
#                     end=temp[8],
#                     strand=temp[9],
#                 )
#                 self.add_protein(
#                     description=temp[13],
#                     external_id=temp[10],
#                     uid=uid,
#                 )
#         log.info(f"Parsed {len(self.assemblies)} assemblies")
#         log.info(f"Parsed {sum([len(i.loci) for i in self.assemblies.values()])} loci")
#         log.info(f"Parsed {len(self.proteins)} non-redundant proteins")


# # accession="BGC0000001|c1|1-1083|+|AEK75490.1|protein_methyltransferase|AEK75490.1"



#!/usr/bin/env python3

import logging
import os
import tempfile
from pathlib import Path

import pandas as pd

from socialgene.base.molbio import Protein
from socialgene.compare_proteins.base import BlastTab, BlastTab_COLUMNS
from socialgene.utils.logging import log
from socialgene.utils.run_subprocess import run_subprocess

# mmseqs must be present on PATH


class MMseqsEasySearch(BlastTab):
    def __init__(self):
        super().__init__()
        self.temp_db_path = None
        self.df = None
        self.name = "MMseqs2 EasySearch"

    @staticmethod
    def make_db(
        fasta_string: str,
        db_path=None,
        **kwargs,
    ):
        """Run MMseqs2 createdb on an input fasta file

        Args:
            fasta_path (str): Path to fasta file
            target_database (str): Path to MMseqs2 database
        """

        command_list = [
            "mmseqs",
            "createdb",
            "stdin",
            str(db_path),
        ]
        mes = run_subprocess(
            command_list=command_list,
            input=fasta_string,
            text=True,
            capture_output=True,
            status=False,
            **kwargs,
        )
        log.debug(mes)
        if not Path(db_path).exists():
            raise FileNotFoundError(
                f"No database from mmseqs make_db found at {db_path}"
            )

    @staticmethod
    def index_database(database_path):
        """Run MMseqs2 createindex on an MMseqs2 database

        Args:
            target_database (str): Path to MMseqs2 database
        """
        with tempfile.TemporaryDirectory() as tmpdirname:
            command_list = [
                "mmseqs",
                "createindex",
                str(database_path),
                str(tmpdirname),
            ]
            mes = run_subprocess(
                command_list=command_list,
                check=False,
                shell=False,
                capture_output=True,
                status=False,
            )
        logging.debug(mes)
        if not os.path.exists(f"{database_path}.idx"):
            logging.warning("No output from MMseqs2 createindex")

    def run(
        self,
        fasta_path,
        target_database,
        cpus=1,
        argstring="",
    ):
        """Search an input fasta file against a target database with external MMseqs2 program

            Args:
                fasta_path (str): Path to fasta file
                target_database (str): Path to MMseqs2 database
                argstring (str): additonal arguments to pass to mmseqs search, as a string
        =
            Returns:
                pandas_df: dataframe of the mmseqs search result
        """
        with tempfile.TemporaryDirectory() as tmpdirname:
            outpath = os.path.join(tmpdirname, "result.m8")
            command_list = [
                "mmseqs",
                "easy-search",
                str(fasta_path),
                str(target_database),
                str(outpath),
                str(tmpdirname),
                "--format-mode",
                "0",
                "--format-output",
                "query,target,pident,alnlen,mismatch,gapopen,qstart,qend,tstart,tend,evalue,bits,qlen,tlen",
                "--threads",
                str(cpus),
            ]
            if argstring:
                argstring = [str(i) for i in argstring.split()]
                command_list.extend(argstring)
            mes = run_subprocess(
                command_list=command_list,
                check=False,
                shell=False,
                capture_output=True,
                status=False,
            )
            logging.debug(mes)
            if not os.path.exists(outpath):
                logging.warning("No output from MMseqs2 search")
            else:
                # sorted for reproducibility
                return (
                    pd.read_csv(
                        outpath,
                        sep="\t",
                        names=BlastTab_COLUMNS,
                        dtype=BlastTab_COLUMNS,
                    )
                    .sort_values(
                        [
                            "query",
                            "score",
                        ],
                        ascending=False,
                    )
                    .reset_index(drop=True)
                )

    def compare_proteins(
        self, queries, targets, cpus=1, argstring="", index=False, **kwargs
    ):
        # loop through protein list and write to temporary fasta file
        if isinstance(queries, Protein):
            queries = [queries]
        if isinstance(targets, Protein):
            targets = [targets]
        if not all([isinstance(i, Protein) for i in queries]):
            raise TypeError("queries must be a list of Protein objects")
        if not all([isinstance(i, Protein) for i in targets]):
            raise TypeError("targets must be a list of Protein objects")

        with tempfile.TemporaryDirectory() as tmpdirname:
            query_fasta_path = Path(tmpdirname, "queries.faa")
            target_dmnd_path = Path(tmpdirname, "targets")
            with open(query_fasta_path, "w") as h:
                for protein in queries:
                    h.write(protein.fasta_string_defline_uid)
            self.make_db(
                fasta_string="".join([i.fasta_string_defline_uid for i in targets]),
                db_path=target_dmnd_path,
            )
            if index:
                self.index_database(target_dmnd_path)
            return self.run(
                fasta_path=query_fasta_path,
                target_database=target_dmnd_path,
                cpus=cpus,
                argstring=argstring,
            )



from socialgene.neo4j.neo4j import GraphDriver
from socialgene.utils.logging import log


def _add_to_neo4j(statement, **kwargs):
    try:
        log.info("\n".join([i.strip() for i in statement.split("\n") if i]))
        summary = (
            GraphDriver().driver.execute_query(
                statement,
                **kwargs,
                database_="neo4j",
            )
        ).summary
        if summary.metadata.get("stats"):
            log.info(summary.metadata.get("stats"))
        else:
            log.info("No properties modified")
    except Exception as e:
        log.warning(e.message)


def _run_transaction_function(statement, **kwargs):
    try:
        log.debug(statement)
        with GraphDriver() as db:
            a = db.run(
                statement,
                **kwargs,
                database_="neo4j",
            ).values()
        return a

    except Exception as e:
        log.error(e)


def add_taxonomic_name_to_assembly(rank):
    _add_to_neo4j(
        f"""
            MATCH (a1:assembly)-[:IS_TAXON|TAXON_PARENT*1..]->(t1:taxid {{rank:"{rank}"}})
            SET a1.{rank} = t1.name;
            """
    )


def add_protein_descriptions():
    log.info(
        "Adding first description to non-redundant proteins (excluding those containing hypothetical); this can take some time."
    )
    log.info("This can take some time")
    res = _run_transaction_function(
        """
                // set protein description as a random input protein's description (except "hypothetical")
                MATCH (p1:protein)
                CALL {
                    WITH p1
                    MATCH (:nucleotide)-[e1:ENCODES]->(p1)
                    WHERE e1.description IS NOT NULL AND NOT toLower(e1.description) CONTAINS "hypothetical"
                    RETURN e1 LIMIT 1
                } IN TRANSACTIONS OF 100 ROWS
                SET p1.description = e1.description
                RETURN count(p1)
            """
    )
    log.info(f"Modified: {res} properties")


def add_antismash_regions_as_edges():
    log.info(
        "Parsing 'import/antismash_results.jsonl' as (:nucleotide)-[e1:ENCODES]->(:proteins) e1.properties"
    )
    _run_transaction_function(
        """
                CALL apoc.load.json("import/antismash_results.jsonl")
                    YIELD value AS jsonl
                CALL {
                    with jsonl
                    UNWIND jsonl.records as record
                    UNWIND keys(record) as nucleotide_id
                    WITH record, nucleotide_id, range(0, size(record[nucleotide_id])-1) as cluster_index
                    UNWIND cluster_index as cluster_index2
                    WITH record, record[nucleotide_id][cluster_index2] as cluster, nucleotide_id,  cluster_index2
                    MATCH (n1:nucleotide {external_id:nucleotide_id})-[c1:ENCODES]->()
                    WHERE cluster.start < (c1.start + 1 ) < cluster.end
                    SET c1.antismash_products = record[nucleotide_id][cluster_index2].products
                    SET c1.antismash_region = cluster_index2
                } IN TRANSACTIONS OF 1000 ROWS;
            """
    )


def add_antismash_regions_as_nodes():
    log.info("Adding new constraint for (:gene_cluster {nuid}), if needed")
    _run_transaction_function(
        """
                CREATE CONSTRAINT gene_cluster IF NOT EXISTS
                FOR (n:gene_cluster)
                REQUIRE (n.uid) IS UNIQUE;
        """
    )
    _run_transaction_function(
        """
                CREATE CONSTRAINT product IF NOT EXISTS
                FOR (n:product)
                REQUIRE (n.uid) IS UNIQUE;
        """
    )
    log.info("Parsing 'import/antismash_results.jsonl' as new antismash nodes")
    _run_transaction_function(
        """
        CALL apoc.load.jsonArray("import/antismash_results.jsonl")
        YIELD value AS jsonl
        UNWIND jsonl.records as record
        UNWIND keys(record) as nucleotide_id
        CALL {
            WITH record, nucleotide_id
        WITH record, nucleotide_id, range(0, size(record[nucleotide_id])-1) as cluster_indices
        UNWIND cluster_indices as cluster_index
        WITH record, record[nucleotide_id][cluster_index] as cluster, nucleotide_id,  cluster_index
        MERGE (a:gene_cluster {uid: nucleotide_id + "_antismash_" + cluster_index})
        SET a:antismash
        SET a += {start: cluster.start, end: cluster.end}
        WITH a, cluster, nucleotide_id
        MATCH (n:nucleotide {external_id: nucleotide_id})
        MERGE (a)<-[:PREDICTED_BGC]-(n)
        WITH a, cluster
           CALL {
              WITH a, cluster
              UNWIND keys(cluster.protoclusters) AS protocluster_key
              WITH a, cluster, cluster.protoclusters[protocluster_key] AS protocluster, protocluster_key
              MERGE (product:product {uid: protocluster.product})
              SET product:antismash
              MERGE (category:category {uid: protocluster.category})
              SET category:antismash
              MERGE (product)-[:IS_A]->(category)
              MERGE (a)-[r:IS_A {tool: protocluster.tool, start: protocluster.start, end: protocluster.end, core_start: protocluster.core_start, core_end: protocluster.core_end}]->(product)
           }
        } IN TRANSACTIONS OF 10000 rows

            """
    )


def delete_antismash_regions_as_nodes():
    _run_transaction_function(
        """
            MATCH (n:gene_cluster)
            WHERE n:antismash
            DETACH DELETE n;
        """
    )


def culture_collections_as_nodes_rels():
    _add_to_neo4j(
        """
                MATCH (a1:assembly) where a1.culture_collection is not null
                WITH a1, split(a1.culture_collection,":")[0] as cc_agency
                MERGE (cc:culture_collection {uid:cc_agency})
                MERGE (a1)-[:FOUND_IN]->(cc);
            """
    )
    _add_to_neo4j(
        """
                MATCH (ccs:culture_collection)
                MATCH (a1:assembly)
                WHERE a1.strain starts with ccs.uid and NOT (a1)-[:FOUND_IN]->(:culture_collection)
                MERGE (a1)-[:FOUND_IN]->(ccs)
            """
    )


def set_mibig_bgc():
    _run_transaction_function(
        """
                MATCH (a1:assembly)<-[:ASSEMBLES_TO]-(n1:nucleotide)
                WHERE a1.uid STARTS WITH "BGC00" AND n1.external_id STARTS WITH "BGC00"
                SET a1:mibig_bgc;
            """
    )


def fix_mibig_taxonomy():
    _add_to_neo4j(
        """
                // Link mibig BGCs to taxa if they aren't
                MATCH (n:assembly)
                where n.uid starts with "BGC00" and not (n)-[:IS_TAXON]->() and n.db_xref is not null
                MATCH (t1:taxid {uid:split(n.db_xref, "taxon:")[1]})
                MERGE (n)-[:IS_TAXON]->(t1);

            """
    )


def antismash_as_separate_nodes():
    log.info("Adding new constrain for (:antismash_bgc {nuid})")
    _run_transaction_function(
        """
                CREATE CONSTRAINT antismash_bgc IF NOT EXISTS
                FOR (n:antismash_bgc)
                REQUIRE (n.uid, n.region) IS UNIQUE;
        """
    )
    log.info("Creating new antismash nodes and linking them to proteins")
    _run_transaction_function(
        """
                MATCH (p1:protein)<-[e1:ENCODES]-(n1:nucleotide)
                WHERE  e1.antismash_region is not null
                WITH n1, e1.antismash_region as ar, min(e1.start) as minstart, max(e1.end) as maxend
                CALL {
                WITH n1, ar, minstart, maxend
                CREATE (bgc:antismash_bgc {uid: n1.uid, region:ar, start:minstart, end:maxend})
                CREATE (n1)-[:PREDICTED_BGC]->(bgc)
                } IN TRANSACTIONS OF 1000 rows;
            """
    )



from itertools import combinations, product
from multiprocessing import Pool

import pandas as pd

from socialgene.compare_proteins.base import HMMDataFrame
from socialgene.compare_proteins.hmm_scoring import mod_score


def picklable_modscore(p1, p2):
    # named tuple doesn't work in multiprocessing
    return mod_score(p1, p2)._asdict()


class CompareDomains(HMMDataFrame):
    def __init__(self):
        self.name = "HMMER annotation comparison with SocialGene"
        self.score_column = "score"

    def compare_proteins(self, queries, targets, **kwargs):
        return pd.DataFrame(
            (
                {
                    "query": i.query.uid,
                    "target": i.target.uid,
                    "query_n_domains": i.query_n_domains,
                    "target_n_domains": i.target_n_domains,
                    "jaccard": i.jaccard,
                    "score": i.mod_score,
                }
                for i in self.compare_many_to_many(
                    queries, targets, filter_non_hits=True
                )
            ),
        ).drop_duplicates(subset=["query", "target"])

    def compare_one_to_one(self, p1, p2):
        return mod_score(p1, p2)

    def compare_one_to_many(self, p1_obj, p2_obj_list, filter_non_hits=True):
        for i in p2_obj_list:
            temp = mod_score(p1_obj, i)
            if not filter_non_hits or temp.jaccard > 0:
                yield temp

    def compare_many_to_many(self, p1_obj_list, p2_obj_list, filter_non_hits=True):
        for i1, i2 in product(p1_obj_list, p2_obj_list):
            temp = mod_score(i1, i2)
            if not filter_non_hits or temp.jaccard > 0:
                yield temp

    def compare_all_to_all(self, p1_obj_list, filter_non_hits=True):
        for i1, i2 in combinations(p1_obj_list, 2):
            temp = mod_score(i1, i2)
            if not filter_non_hits or temp.jaccard > 0:
                yield temp

    def compare_all_to_all_parallel(self, p1_obj_list, cpus=1, only_hits=True):
        # have to use _calculate_mod_score_not_named because named tuple can't pickle "protein_comparison_modscore"
        with Pool(cpus) as p:
            for i in p.starmap(
                mod_score,
                combinations(p1_obj_list, 2),
            ):
                if not only_hits or i.jaccard > 0.001:
                    yield i



from functools import partial
from io import StringIO
from pathlib import Path
from uuid import uuid4

from Bio import SeqIO

import socialgene.utils.file_handling as fh
from socialgene.utils.logging import log

fasta_prefixes = {
    "swissprot": "sp|",
    "trembl": "tr|",
    # "uniclust1": "uc",
    # "uniclust2": "cl|",
    # "genbank": "gb|",
    # "ncbi_ref": "ref|",
    # "patents": "pat|",
    # "ncbi_gi": "gi|",
}


def get_after_pipe(x, n=1):
    return x.split("|")[n]


fasta_prefixes_get_identifier = {
    "swissprot": partial(get_after_pipe, n=1),
    "trembl": partial(get_after_pipe, n=1),
}


def check_prefix(id):
    for k, v in fasta_prefixes.items():
        if id.startswith(v):
            return k
    return None


class FastaParserMixin:
    def __init__(self):
        pass

    @staticmethod
    def parse_id(id):
        try:
            prefix_key = check_prefix(id)
            if prefix_key:
                return fasta_prefixes_get_identifier[prefix_key](id)
        except Exception:
            return id
        return id

    # havent tested/worked on this since lots of updates
    def parse_fasta_file(
        self,
        input: str = None,
        defline_magic: bool = False,
    ):
        """Parse a protein fasta file

        Args:
            input (str): path to fasta file
            defline_magic (bool): Parse out fasta deflines with defined structures, currently only works for uniprot (deflines that begin with: sp| or tr|)
        """
        # if not appending, reset self.records
        if isinstance(input, str) and input.startswith(">"):
            assembly_id = str(uuid4())
            _open = StringIO(input)
            input_from = "input string"
        else:
            input = Path(input)
            try:
                assembly_id = input.stem
            except Exception:
                assembly_id = str(uuid4())
            _open = fh.open_read(input)
            input_from = input
        with _open as handle:
            self.add_assembly(uid=assembly_id, parent=self)
            self.assemblies[assembly_id].add_locus(external_id=assembly_id)
            record_counter = 0
            for seq_record in SeqIO.parse(handle, "fasta"):
                # probably could use some more defensive programming here
                if defline_magic:
                    id = self.parse_id(seq_record.id)
                else:
                    id = seq_record.id
                uid = self.add_protein(
                    description=seq_record.description,
                    external_id=id,
                    sequence=str(seq_record.seq),
                    return_uid=True,
                )
                self.assemblies[assembly_id].loci[assembly_id].add_feature(
                    type="protein",
                    description=seq_record.description,
                    external_id=id,
                    uid=uid,
                    start=record_counter,
                    end=record_counter,
                    strand=0,
                )
                record_counter += 1
        log.info(f"Read {record_counter} proteins from {input_from}")



# from socialgene.base.socialgene import SocialGene
# from socialgene.neo4j.neo4j import GraphDriver  # grab the the neo4j connection


# a = SocialGene()

# with GraphDriver() as db:
#     results = db.run(
#         """
# MATCH (p1:protein) RETURN p1.uid
# """
#     )
#     for i in results:
#         i = i[0]
#         _ = a.add_protein(uid=i)


# with GraphDriver() as db:
#     results = db.run(
#         """
# MATCH (a1:assembly) RETURN a1
# """
#     )
#     for i in results:
#         i = i[0]
#         a.add_assembly(id=i._properties["id"])
#         a.assemblies[i._properties["id"]].info = (
#             a.assemblies[i._properties["id"]].info | i._properties
#         )


# with GraphDriver() as db:
#     results = db.run(
#         """
# MATCH (n1:nucleotide)-[:ASSEMBLES_TO]-(a1:assembly) RETURN {n:n1, a:a1}
# """
#     )
#     for i in results:
#         i = i[0]
#         a.assemblies[i["a"]._properties["id"]].add_locus(
#             i["n"]._properties["internal_id"]
#         )
#         a.assemblies[i["a"]._properties["id"]].loci[
#             i["n"]._properties["internal_id"]
#         ].info = (
#             a.assemblies[i["a"]._properties["id"]]
#             .loci[i["n"]._properties["internal_id"]]
#             .info
#             | i["n"]._properties
#         )


# with GraphDriver() as db:
#     results = db.run(
#         """
# MATCH (a1:assembly)<-[:ASSEMBLES_TO]-(n1:nucleotide)-[e1:ENCODES]->(p1:protein) RETURN {a:a1.uid, n:n1.internal_id, p:p1.uid, start:e1.start, end:e1.end, strand:e1.strand}
# """
#     )
#     for i in results:
#         i = i[0]
#         a.assemblies[i["a"]].loci[i["n"]].add_feature(
#             type="protein",
#             id=i["p"],
#             start=i["start"],
#             end=i["end"],
#             strand=i["strand"],
#         )


# with GraphDriver() as db:
#     results = db.run(
#         """
# MATCH (p1:protein)<-[a1:ANNOTATES]-(h1:hmm) RETURN {p:p1.uid, a:a1, h:h1.uid}
# """
#     )
#     for i in results:
#         i = i[0]
#         z = {"hmm_id": i["h"]} | i["a"]._properties
#         a.proteins[i["p"]].add_domain(**z, exponentialized=False)

# with GraphDriver() as db:
#     results = db.run(
#         """
# MATCH (p1)-[b:BLASTP]->(p2:protein)
# RETURN
#     p1.uid as query,
#     p2.uid as subject,
#     b.bitscore as bitscore,
#     b.evalue as evalue,
#     b.gapopen as gapopen,
#     b.length as length,
#     b.mismatch as mismatch,
#     b.pident as pident,
#     b.qcovhsp as qcovhsp,
#     b.qend as qend,
#     b.qstart as qstart,
#     b.send as send,
#     b.sstart as sstart
# """
#     )
#     a.blastp_df = results.to_df()


# with GraphDriver() as db:
#     results = db.run(
#         """
# MATCH (p1)-[b:MMSEQS2]->(p2:protein)
# RETURN
#     p1.uid as cluster_representative,
#     p2.uid as external_id
# """
#     )
#     a.mmseqs2_df = results.to_df()


# with GraphDriver() as db:
#     results = db.run(
#         """
# MATCH (p1)-[b:SIMILAR]-(p2:protein)
# RETURN DISTINCT
#     p1.uid as p1,
#     p2.uid as p2,
#     b.score as score
# """
#     )
#     a.sim_df = results.to_df()



import requests
from rich.progress import (
    BarColumn,
    DownloadColumn,
    Progress,
    TextColumn,
    TimeRemainingColumn,
    TransferSpeedColumn,
)

progress = Progress(
    TextColumn("[bold blue]{task.fields[filename]}", justify="right"),
    BarColumn(bar_width=None),
    "[progress.percentage]{task.percentage:>3.1f}%",
    "•",
    DownloadColumn(),
    "•",
    TransferSpeedColumn(),
    "•",
    TimeRemainingColumn(),
    transient=True,
)


def download(url, filepath):
    res = requests.get(
        "https://www.npatlas.org/static/downloads/NPAtlas_download.json", stream=True
    )
    if "Content-length" in res.headers:
        total = int(res.headers["Content-length"])
    else:
        total = 0
    tt = 0
    with progress:
        task_id = progress.add_task(
            "download", filename="NPAtlas_download.json", total=total
        )
        with open(filepath, "wb") as f:
            for i in res.iter_content(8192):
                _ = f.write(i)
                progress.update(task_id, advance=len(i))
                tt += len(i)



import argparse
from pathlib import Path

from socialgene.parsers.ncbi_taxonomy import (
    merge_taxonomy,
    process_names_dmp,
    process_nodes_dmp,
)
from socialgene.utils.untargz import untargz

# tax_url = "ftp://ftp.ncbi.nih.gov/pub/taxonomy/taxdump.tar.gz"
# tax_dump_hash_url = "ftp://ftp.ncbi.nih.gov/pub/taxonomy/taxdump.tar.gz.md5"


parser = argparse.ArgumentParser(description="Parse NcbiAssemblies taxonomy")
parser.add_argument(
    "--taxdump_path",
    help="path to ncbi's taxdump.tar.gz",
    required=True,
)


def write_tsv(input_data, filepath, columns=None, mode="w"):
    input_data.to_csv(
        path_or_buf=filepath,
        sep="\t",
        na_rep="",
        float_format=None,
        columns=columns,
        header=False,
        index=False,
        index_label=None,
        mode=mode,
        encoding=None,
        compression="infer",
        quoting=None,
        quotechar='"',
        lineterminator=None,
        chunksize=None,
        date_format=None,
        doublequote=True,
        escapechar=None,
        decimal=".",
        errors="strict",
        storage_options=None,
    )


def main():
    args = parser.parse_args()
    taxdump_path = Path(args.taxdump_path)  # taxdump.tar.gz

    names_handle = untargz(taxdump_path, file_name="names.dmp")
    nodes_handle = untargz(taxdump_path, file_name="nodes.dmp")
    names_df = process_names_dmp(file_handle=names_handle)
    nodes_df = process_nodes_dmp(file_handle=nodes_handle)

    write_tsv(
        input_data=merge_taxonomy(names_df=names_df, nodes_df=nodes_df),
        filepath="nodes_taxid",
    )

    write_tsv(input_data=nodes_df.iloc[:, 0:2], filepath="taxid_to_taxid")


if __name__ == "__main__":
    main()



import re
from typing import List

from socialgene.addons.chemistry.nr import ChemicalCompoundNode
from socialgene.addons.npclassifier.nr import (
    NPClassifierClass,
    NPClassifierPathway,
    NPClassifierSuperclass,
)
from socialgene.neo4j.neo4j_element import Node, Relationship


class GnpsLibrarySpectrumNode(Node):
    neo4j_label = ["gnps_library_spectrum"]
    description = "Represents a GNPS library spectrum"
    uid = ["uid"]
    constraints_unique = ["uid"]
    required_properties = ["uid"]
    property_specification = {
        "uid": str,
        "compound_name": str,
        "compound_source": str,
        "pi": str,
        "data_collector": str,
        "adduct": str,
        "precursor_mz": float,
        "exactmass": float,
        "charge": int,
        "cas_number": str,
        "pubmed_id": str,
        "smiles": str,
        "inchi": str,
        "inchi_aux": str,
        "library_class": str,
        "ionmode": str,
        "libraryqualitystring": str,
        "mqscore": float,
        "tic_query": float,
        "rt_query": float,
        "mzerrorppm": float,
        "sharedpeaks": int,
        "massdiff": float,
        "libmz": float,
        "specmz": float,
        "speccharge": int,
        "moleculeexplorerdatasets": str,
        "moleculeexplorerfiles": str,
        "molecular_formula": str,
        "inchikey": str,
        "inchikey_planar": str,
    }

    @staticmethod
    def _extract_all_CCMSLIB(x) -> List:
        return re.findall("CCMSLIB[0-9]{11}", x)


class IonSourceNode(Node):
    neo4j_label = ["ion_source"]
    description = "Represents an ion source"
    property_specification = {
        "uid": str,
    }
    constraints_unique = ["uid"]


class InstrumentNode(Node):
    neo4j_label = ["instrument"]
    description = "Represents an instrument"
    property_specification = {
        "uid": str,
    }
    constraints_unique = ["uid"]


class OrganismNode(Node):
    neo4j_label = ["gnps_organism"]
    description = "Represents an organism (as defined by GNPS)"
    property_specification = {
        "uid": str,
    }
    constraints_unique = ["uid"]


class FromIonRel(Relationship):
    neo4j_label = "FROM"
    description = "Connects a GNPS spectrum to an ion source"
    start_class = GnpsLibrarySpectrumNode
    end_class = IonSourceNode


class FromInstrumentRel(Relationship):
    neo4j_label = "FROM"
    description = "Connects a GNPS spectrum to an instrument source"
    start_class = GnpsLibrarySpectrumNode
    end_class = InstrumentNode


class FromOrganismRel(Relationship):
    neo4j_label = "FROM"
    description = "Connects a GNPS spectrum to an organism (as defined by GNPS)"
    start_class = GnpsLibrarySpectrumNode
    end_class = OrganismNode


class GnpsLibrarySpectrumToNPClassifierClass(Relationship):
    neo4j_label = "IS_A"
    description = "Represents a relationship between gnps_library_spectrum nodes"
    start_class = GnpsLibrarySpectrumNode
    end_class = NPClassifierClass


class GnpsLibrarySpectrumToNPClassifierSuperclass(Relationship):
    neo4j_label = "IS_A"
    description = "Represents a relationship between gnps_library_spectrum nodes"
    start_class = GnpsLibrarySpectrumNode
    end_class = NPClassifierSuperclass


class GnpsLibrarySpectrumToNPClassifierPathway(Relationship):
    neo4j_label = "IS_A"
    description = "Represents a relationship between gnps_library_spectrum nodes"
    start_class = GnpsLibrarySpectrumNode
    end_class = NPClassifierPathway


class GnpsLibraryToChem(Relationship):
    neo4j_label = "IS_A"
    description = "Connects a GNPS library spectrum to a chemical compound"
    start_class = GnpsLibrarySpectrumNode
    end_class = ChemicalCompoundNode



import tarfile


def untargz(input_path, file_name=False, file_name_ends_with=False):
    """Get a file from a tar.gz
    Parameters
    ----------
    input_path : str
        path to the tar.gz
    file_name : str
        file name
    file_name_ends_with : str
        file ned with
    Returns
    -------
    io.BufferedReader object
    Raises
    -------
    ValueError
        "Too many matches. Function should only pull out a single file"
    """
    # extract domtblout from the tar
    tarfile_object = tarfile.open(input_path)
    temp = tarfile_object.getnames()
    if file_name and file_name_ends_with:
        raise ValueError("Must only set one of 'file_name' or 'file_name_ends_with'")
    if file_name:
        match_result = [x == file_name for x in temp]
        match_result = [i for i, x in enumerate(match_result) if x]
        if len(match_result) > 1:
            raise ValueError(
                "Too many matches. Function should only pull out a single file"
            )
    elif file_name_ends_with:
        # endswith matching because domtblout files in the tar.gz
        # have names that are hashes
        match_result = [x.endswith(file_name_ends_with) for x in temp]
        match_result = [i for i, x in enumerate(match_result) if x]
        if len(match_result) > 1:
            raise ValueError(
                "Too many matches. Function should only pull out a single file"
            )
    else:
        raise ValueError("Must set either 'file_name' or 'file_name_ends_with'")
    file_name_to_extract = temp[match_result[0]]
    file = tarfile_object.extractfile(file_name_to_extract)
    return file



import argparse
import csv
import platform

from socialgene.config import env_vars
from socialgene.utils.logging import log

parser = argparse.ArgumentParser(
    description="Export common parameters, especially for nextflow/neo4j"
)

parser.add_argument(
    "--outpath",
    help="Output filepath",
    default="socialgene_parameters",
)

# expect a string "None"; not None
parser.add_argument("--genome_download_command", required=False, default="None")

fields = {
    "SG_LOC_NEO4J": None,
    "SG_LOC_HMMS": None,
    "NEO4J_dbms_memory_pagecache_size": None,
    "NEO4J_dbms_memory_heap_initial__size": None,
    "NEO4J_dbms_memory_heap_max__size": None,
    "HMMSEARCH_IEVALUE": None,
    "HMMSEARCH_BACKGROUND": None,
    "HMMSEARCH_BIASFILTER": None,
    "HMMSEARCH_NULL2": None,
    "HMMSEARCH_SEED": None,
    "HMMSEARCH_Z": None,
    "HMMSEARCH_DOMZ": None,
    "HMMSEARCH_F1": None,
    "HMMSEARCH_F2": None,
    "HMMSEARCH_F3": None,
    "HMMSEARCH_E": None,
    "HMMSEARCH_DOME": None,
    "HMMSEARCH_INCE": None,
    "HMMSEARCH_INCDOME": None,
    "HMMSEARCH_BITCUTOFFS": None,
}


def main():
    args = parser.parse_args()

    save_list = ["socialgene_parameter_export"]

    with open(args.outpath, "a") as f:
        for k, v in env_vars.items():
            if k in fields:
                fields[k] = v
            else:
                log.info(f"Unexpected parameter: '{k}'")
        for v in fields.values():
            save_list.append(v)

        # TODO: trycatch all this and check length at end
        save_list.append(platform.sys.platform)
        save_list.append(" ".join(platform.architecture()))
        save_list.append(platform.sys.executable)
        save_list.append(platform.sys.version)
        save_list.append(args.genome_download_command)
        save_list = [str(i) for i in save_list]
        save_list = [i.strip('"') for i in save_list]
        save_list = [i.replace("\n", "") for i in save_list]
        save_list = [i.replace("\t", "") for i in save_list]

        writer = csv.writer(
            f,
            quoting=csv.QUOTE_MINIMAL,
            delimiter="\t",
        )
        writer.writerows([save_list])


# TODO:
# date/time
# Py/Next/Neo versions
# SG version and/or git sha
# domputer info
# source genome info


if __name__ == "__main__":
    main()



"""https://www.ebi.ac.uk/chembl"""

from socialgene.neo4j.neo4j_element import Node


class CHEMBL(Node):
    neo4j_label = ["chembl"]
    description = "Represents a CHEMBL term"
    property_specification = {
        "uid": str,
    }
    constraints_unique = ["uid"]



# from socialgene.neo4j.neo4j_element import Node, Relationship
# try to use classes already created that inherit from Nodes() or Relationship()
# to do that create a mixin class in a separate file and then import that



from pathlib import Path

from socialgene.addons.chebi.nr import ChebiNode
from socialgene.addons.chemistry.nr import ChemicalCompoundNode
from socialgene.addons.classyfire.nr import ClassyFireNode
from socialgene.addons.gnps_library.nr import GnpsLibrarySpectrumNode
from socialgene.addons.mibig.nr import Mibig
from socialgene.addons.npatlas.nr import (
    MibigToNPAtlas,
    NPAtlasNode,
    NPAtlasPublication,
    NPAtlasToChebi,
    NPAtlasToChem,
    NPAtlasToClassyFireAlternativeParents,
    NPAtlasToClassyFireDirectParent,
    NPAtlasToClassyFireIntermediateNodes,
    NPAtlasToClassyFireLowestClass,
    NPAtlasToGnps,
    NPAtlasToNpclassifierClass,
    NPAtlasToNpclassifierPathway,
    NPAtlasToNpclassifierSuperclass,
    NPAtlasToPublication,
    NPAtlasToTaxID,
)
from socialgene.addons.npclassifier.nr import (
    NPClassifierClass,
    NPClassifierPathway,
    NPClassifierSuperclass,
)
from socialgene.addons.npmrd.nr import Npmrd
from socialgene.base.chem import ChemicalCompound
from socialgene.nextflow.nodes import TAXID
from socialgene.utils.download import download as downloader
from socialgene.utils.logging import log

NPATALAS_URL = "https://www.npatlas.org/static/downloads/NPAtlas_download.json"


def _download_npatlas(outpath, url=NPATALAS_URL):
    # check if exist and not empty
    if Path(outpath).exists() and Path(outpath).stat().st_size > 0:
        log.debug(f"File already exists at {outpath}")
    else:
        downloader(url, outpath)


class NPAtlasEntry:
    __slots__ = [
        "entry",
        "uid",
        "original_name",
        "mol_formula",
        "mol_weight",
        "exact_mass",
        "inchikey",
        "smiles",
        "cluster_id",
        "node_id",
        "synonyms",
        "inchi",
        "m_plus_h",
        "m_plus_na",
        "origin_reference",
        "external_ids",
        "ncbi_taxid",
        "genus",
        "species",
        "lowest_classyfire",
        "gnps_ids",
        "mibig_ids",
        "npmrd_ids",
        "predicted_chebi_terms",
        "classyfire_direct_parent",
        "classyfire_intermediate_nodes",
        "classyfire_alternative_parents",
        "npclassifier_classes",
        "npclassifier_pathways",
        "npclassifier_superclasses",
        "npclassifier_glycoside",
    ]

    def __init__(self, entry) -> None:
        self.entry = entry
        self.uid = None
        self.original_name = None
        self.mol_formula = None
        self.mol_weight = None
        self.exact_mass = None
        self.inchikey = None
        self.smiles = None
        self.cluster_id = None
        self.node_id = None
        self.synonyms = None
        self.inchi = None
        self.m_plus_h = None
        self.m_plus_na = None
        self.origin_reference = None
        self.external_ids = None
        self.ncbi_taxid = None
        self.genus = None
        self.species = None
        self.lowest_classyfire = None
        self.gnps_ids = set()
        self.mibig_ids = set()
        self.npmrd_ids = set()
        self.predicted_chebi_terms = {}
        self.classyfire_direct_parent = None
        self.classyfire_intermediate_nodes = None
        self.classyfire_alternative_parents = None
        self.npclassifier_classes = None
        self.npclassifier_pathways = None
        self.npclassifier_superclasses = None
        self.npclassifier_glycoside = None

    def parse(self):
        self._parse_single_entry()
        self._assign_lowest_classyfire()
        self._assign_classyfire_direct_parent()
        self._assign_classyfire_intermediate_nodes()
        self._assign_classyfire_alternative_parents()
        self._assign_npclassifier()
        self._assign_publication()

    def _parse_single_entry(self):
        self.uid = self.entry.get("npaid", None)
        self.original_name = self.entry.get("original_name", None)
        self.mol_formula = self.entry.get("mol_formula", None)
        self.mol_weight = float(self.entry.get("mol_weight", None))
        self.exact_mass = float(self.entry.get("exact_mass", None))
        self.inchikey = str(self.entry.get("inchikey", None))
        self.smiles = str(self.entry.get("smiles", None))
        self.cluster_id = int(self.entry.get("cluster_id", None))
        self.node_id = int(self.entry.get("node_id", None))
        self.inchi = self.entry.get("inchi", None)
        self.m_plus_h = float(self.entry.get("m_plus_h", None))
        self.m_plus_na = float(self.entry.get("m_plus_na", None))
        self.origin_reference = self.entry.get("origin_reference", None)
        self.synonyms = None
        self._assign_taxon()
        self._assign_external_ids()
        self._assign_chebi_terms()

    def _assign_publication(self):
        self.origin_reference = NPAtlasPublication()
        try:
            self.origin_reference.properties["doi"] = (
                self.origin_reference._extract_doi(
                    self.entry.get("origin_reference").get("doi", None)
                )
            )
        except Exception as e:
            log.debug(f"Failed to extract doi: {e}")

        for i in ["pmid", "authors", "title", "journal", "year"]:
            try:
                self.origin_reference.properties[i] = self.entry.get(
                    "origin_reference"
                ).get(i, None)
            except Exception as e:
                log.debug(f"Failed to extract {i}: {e}")

    def _assign_lowest_classyfire(self):
        # not all entries have classyfire data at every level, just assign the most specific one
        for i in ["kingdom", "superclass", "class", "subclass"]:
            try:
                self.lowest_classyfire = self.entry["classyfire"][i][
                    "chemont_id"
                ].removeprefix("CHEMONTID:")
            except Exception as e:
                log.debug(e)

    def _assign_classyfire_direct_parent(self):
        try:
            self.classyfire_direct_parent = self.entry["classyfire"]["direct_parent"][
                "chemont_id"
            ].removeprefix("CHEMONTID:")
            self.classyfire_direct_parent = int(self.classyfire_direct_parent)
        except Exception as e:
            log.debug(e)

    def _assign_classyfire_intermediate_nodes(self):
        try:
            self.classyfire_intermediate_nodes = [
                int(x["chemont_id"].removeprefix("CHEMONTID:"))
                for x in self.entry["classyfire"]["intermediate_nodes"]
            ]
        except Exception as e:
            log.debug(e)

    def _assign_classyfire_alternative_parents(self):
        try:
            self.classyfire_alternative_parents = [
                int(x["chemont_id"].removeprefix("CHEMONTID:"))
                for x in self.entry["classyfire"]["alternative_parents"]
            ]
        except Exception as e:
            log.debug(e)

    def _assign_npclassifier(self):
        try:
            self.npclassifier_classes = self.entry["npclassifier"]["class_results"]
        except Exception as e:
            log.debug(e)
        try:
            self.npclassifier_pathways = self.entry["npclassifier"]["pathway_results"]
        except Exception as e:
            log.debug(e)
        try:
            self.npclassifier_superclasses = self.entry["npclassifier"][
                "superclass_results"
            ]
        except Exception as e:
            log.debug(e)
        try:
            if self.entry["npclassifier"]["isglycoside"]:
                self.npclassifier_glycoside = True
        except Exception as e:
            log.debug(e)

    def _assign_taxon(self):
        try:
            self.ncbi_taxid = self.entry["origin_organism"]["taxon"]["ncbi_id"]
        except Exception:
            pass
        try:
            self.genus = self.entry["origin_organism"]["genus"]
        except Exception:
            pass
        try:
            self.species = self.entry["origin_organism"]["species"]
        except Exception:
            pass

    def _assign_chebi_terms(self):
        try:
            for term in self.entry["classyfire"]["predicted_chebi_terms"]:
                # extract name and uid as int from strings like "hydroxycoumarin (CHEBI:37912)"
                uid = term.split(" ")[-1].removeprefix("(CHEBI:").removesuffix(")")
                name = term.split("(")[0]
                uid = uid.strip()
                name = name.strip()
                self.predicted_chebi_terms[uid] = name
        except Exception as e:
            log.debug(e)

    def _assign_external_ids(self):
        for external_id in self.entry["external_ids"]:
            try:
                match external_id["external_db_name"]:
                    case "gnps":
                        # get first becasue some npaid have multiple gnps ids, eg:
                        # "CCMSLIB00005722620%NCGC00179860-02!6-[(3R,4S,5S,7R)-7-[(2S,3S,5S)-5-ethyl-5-[(2R,5R,6S)-5-ethyl-5-hydroxy-6-methyloxan-2-yl]-3-methyloxolan-2-yl]-4-hydroxy-3,5-dimethyl-6-oxononyl]-2-hydroxy-3-methylbenzoic acid [IIN-based on: CCMSLIB00000848016]%3""
                        uid = GnpsLibrarySpectrumNode._extract_all_CCMSLIB(
                            external_id["external_db_code"]
                        )[0]
                        self.gnps_ids.add(
                            GnpsLibrarySpectrumNode(properties={"uid": uid})
                        )
                    case "npmrd":
                        self.npmrd.add(
                            Npmrd(properties={"uid": external_id["external_db_code"]})
                        )
                    case "mibig":
                        self.mibig_ids.add(
                            Mibig(properties={"uid": external_id["external_db_code"]})
                        )
                    case _:
                        pass
            except Exception as e:
                log.debug(e)

    @property
    def node(self):
        return NPAtlasNode(
            properties={
                "uid": self.uid,
                "original_name": self.original_name,
                "mol_formula": self.mol_formula,
                "mol_weight": self.mol_weight,
                "exact_mass": self.exact_mass,
                "inchikey": self.inchikey,
                "smiles": self.smiles,
                "cluster_id": self.cluster_id,
                "node_id": self.node_id,
                "synonyms": self.synonyms,
                "inchi": self.inchi,
                "m_plus_h": self.m_plus_h,
                "m_plus_na": self.m_plus_na,
                "genus": self.genus,
                "species": self.species,
            }
        )

    def link_publication(self) -> set:
        if self.origin_reference is None:
            return set()
        return {NPAtlasToPublication(start=self.node, end=self.origin_reference)}

    def link_ncbi_taxid(self) -> set:
        if self.ncbi_taxid is None:
            return set()
        return {
            NPAtlasToTaxID(
                start=TAXID(properties={"uid": str(self.ncbi_taxid)}), end=self.node
            )
        }

    def link_gnps(self) -> set:
        if len(self.gnps_ids) == 0:
            return set()
        return {NPAtlasToGnps(start=self.node, end=x) for x in self.gnps_ids}

    def link_mibig(self) -> set:
        if len(self.mibig_ids) == 0:
            return set()
        return {MibigToNPAtlas(start=x, end=self.node) for x in self.mibig_ids}

    def link_classyfire(self) -> set:
        if self.lowest_classyfire is None:
            return set()
        return {
            NPAtlasToClassyFireLowestClass(
                start=self.node,
                end=ClassyFireNode(properties={"uid": int(self.lowest_classyfire)}),
            )
        }

    def link_classyfire_direct_parent(self) -> set:
        if self.classyfire_direct_parent is None:
            return set()
        return {
            NPAtlasToClassyFireDirectParent(
                start=self.node,
                end=ClassyFireNode(
                    properties={"uid": int(self.classyfire_direct_parent)}
                ),
            )
        }

    def link_classyfire_intermediate_nodes(self) -> set:
        if self.classyfire_intermediate_nodes is None:
            return set()
        return {
            NPAtlasToClassyFireIntermediateNodes(
                start=self.node, end=ClassyFireNode(properties={"uid": int(x)})
            )
            for x in self.classyfire_intermediate_nodes
        }

    def link_classyfire_alternative_parents(self) -> set:
        if self.classyfire_alternative_parents is None:
            return set()
        return {
            NPAtlasToClassyFireAlternativeParents(
                start=self.node, end=ClassyFireNode(properties={"uid": int(x)})
            )
            for x in self.classyfire_alternative_parents
        }

    def link_npclassifier_classes(self) -> set:
        if self.npclassifier_classes is None:
            return set()
        return {
            NPAtlasToNpclassifierClass(
                start=self.node, end=NPClassifierClass(properties={"uid": x})
            )
            for x in self.npclassifier_classes
        }

    def link_npclassifier_pathways(self) -> set:
        if self.npclassifier_pathways is None:
            return set()
        return {
            NPAtlasToNpclassifierPathway(
                start=self.node, end=NPClassifierPathway(properties={"uid": x})
            )
            for x in self.npclassifier_pathways
        }

    def link_npclassifier_superclasses(self) -> set:
        if self.npclassifier_superclasses is None:
            return set()
        return {
            NPAtlasToNpclassifierSuperclass(
                start=self.node, end=NPClassifierSuperclass(properties={"uid": x})
            )
            for x in self.npclassifier_superclasses
        }

    def link_chem(self) -> set:
        try:
            chem = None
            if self.inchi:
                chem = ChemicalCompound(self.inchi)
            elif self.smiles:
                chem = ChemicalCompound(self.smiles)
            if chem:
                a = ChemicalCompoundNode()
                a.fill_from_dict(chem.base_properties | chem.hash_dict)
                return {NPAtlasToChem(self.node, a)}
        except Exception:
            return set()

    def link_chebi(self) -> set:
        try:
            return {
                NPAtlasToChebi(
                    start=self.node,
                    end=ChebiNode(properties={"uid": int(k), "name": v}),
                )
                for k, v in self.predicted_chebi_terms.items()
            }
        except Exception:
            return set()

    def get_links(self):
        return {
            NPAtlasToPublication: self.link_publication(),
            NPAtlasToTaxID: self.link_ncbi_taxid(),
            NPAtlasToGnps: self.link_gnps(),
            MibigToNPAtlas: self.link_mibig(),
            NPAtlasToClassyFireLowestClass: self.link_classyfire(),
            NPAtlasToClassyFireDirectParent: self.link_classyfire_direct_parent(),
            NPAtlasToClassyFireIntermediateNodes: self.link_classyfire_intermediate_nodes(),
            NPAtlasToClassyFireAlternativeParents: self.link_classyfire_alternative_parents(),
            NPAtlasToNpclassifierClass: self.link_npclassifier_classes(),
            NPAtlasToNpclassifierPathway: self.link_npclassifier_pathways(),
            NPAtlasToNpclassifierSuperclass: self.link_npclassifier_superclasses(),
            NPAtlasToChem: self.link_chem(),
            NPAtlasToChebi: self.link_chebi(),
        }



import re
import zlib
from collections import Counter
from io import StringIO
from pathlib import Path
from typing import Dict, List
from uuid import uuid4

from Bio import Seq, SeqIO

import socialgene.utils.file_handling as fh
from socialgene.utils.logging import log


def get_seqio_start(seq_feature):
    return seq_feature.location.start.real + 1


def get_seqio_end(seq_feature):
    return seq_feature.location.end.real


class GenbankParserMixin:
    def __init__(self):
        self._self._count_loci_in_file
        pass

    def _add_assembly(self, input_path, seq_record):
        """
        The function `_add_assembly` adds an assembly ID to a sequence record, using either the dbxrefs
        attribute, the input file name, or a random unique ID.

        Args:
          input_path: The input_path parameter is the path to the input file. It is used to determine
        the assembly ID if it is not provided in the seq_record object.
          seq_record: The `seq_record` parameter is an object that represents a sequence record. It
        likely contains information about a biological sequence, such as DNA or protein, including its
        sequence data and associated metadata.

        Returns:
          the assembly ID.
        """
        if hasattr(seq_record, "dbxrefs"):
            # try to retrieve assembly id from the seqrecord dbxref
            assembly_id = self._extract_from_dbxrefs(
                input_list=getattr(seq_record, "dbxrefs", None), prefix="Assembly:"
            )
        if not assembly_id:
            # use filename as assembly id
            assembly_id = str(Path(str(input_path).removesuffix(".gz")).stem)
        if not assembly_id:
            # use random unique id as assembly id
            assembly_id = str(uuid4())
        self.add_assembly(uid=assembly_id, parent=self)
        return assembly_id

    def _add_locus(self, seq_record, assembly_id):
        """
        The function adds a locus object to an assembly object and returns the ID of the added sequence
        record.

        Args:
          seq_record: The `seq_record` parameter is a sequence record object. It typically contains
        information about a biological sequence, such as its ID, sequence data, and annotations. In this
        context, it is being used to add a locus object to an assembly object.
          assembly_id: The `assembly_id` parameter is the identifier of the assembly object to which the
        locus object will be added.

        Returns:
          the ID of the sequence record that was added to the assembly object.
        """
        # add a locus object to an assembly object
        self.assemblies[assembly_id].add_locus(external_id=seq_record.id)
        return seq_record.id

    def _add_taxon(self, seq_feature, assembly_id):
        """
        The function `_add_taxon` extracts the taxon ID from a sequence feature's qualifiers and assigns
        it to the corresponding assembly ID in a dictionary.

        Args:
          seq_feature: The `seq_feature` parameter is a sequence feature object that represents a
        specific feature of a DNA or protein sequence, such as a gene or a coding region.
          assembly_id: The `assembly_id` parameter is an identifier for a specific assembly. It is used
        to access and modify the `taxid` attribute of the assembly object in the `self.assemblies`
        dictionary.
        """
        try:
            self.assemblies[assembly_id].taxid = self._extract_from_dbxrefs(
                input_list=seq_feature.qualifiers.get("db_xref"), prefix="taxon:"
            )
            log.debug(f"{assembly_id} is taxon: {self.assemblies[assembly_id].taxid}")
        except Exception as e:
            log.debug(e)

    def _process_feature_note_protein_exceptions(self, note: str) -> Dict[str, bool]:
        """
        Processes a given feature note and returns a dictionary indicating whether certain protein exceptions are present.

        Args:
          note (str): The `note` parameter is a string that contains information about a locus feature.

        Returns:
          a dictionary where the keys are the keys from the `bad_proteins` dictionary and the values are
        the string "true" for each key that matches a pattern in the `note` string.
        """
        bad_proteins = {
            "partial_on_complete_genome": "partial on complete genome",
            "missing_start": "missing start",
            "missing_stop": "missing stop",
            "internal_stop": "internal stop",
            "partial_in_the_middle_of_a_contig": "partial in the middle of a contig",
            "missing_N_terminus": "missing N-terminus",
            "missing_C_terminus": "missing C-terminus",
            "frameshifted": "frameshifted",
            "too_short_partial_abutting_assembly_gap": "too short partial abutting assembly gap",
            "incomplete": "incomplete",
        }
        # The exact string "true" is looked for by neo4j admin import, not True
        # https://neo4j.com/docs/operations-manual/current/tools/neo4j-admin/neo4j-admin-import/#import-tool-header-format-properties
        return {k: "true" for k, v in bad_proteins.items() if re.search(v, note)}

    def _process_go(self, note: str) -> Dict[str, List[str]]:
        """
        Processes a given feature note and returns a dictionary with a single
        key "goterms" whose value is a list of strings representing GO terms found in the input string.

        Args:
          note (str): The `note` parameter is a string that contains information about a locus feature.

        Returns:
          a dictionary with a single key "goterms" and a value that is a list of strings. The strings in
        the list are extracted from the input string "note" using regular expression pattern matching.
        The regular expression pattern "GO:[0-9]{7}" matches strings that start with "GO:" followed by
        exactly 7 digits.
        """
        return {"goterms": re.findall("GO:[0-9]{7}", note)}

    def _process_feature_note(self, seq_feature) -> Dict:
        """
        The function `_process_feature_note` processes the note qualifiers of a sequence feature and
        returns a dictionary containing information extracted from the note.

        Args:
          seq_feature: The `seq_feature` parameter is a SeqFeature object that represents a feature
        (e.g., gene, CDS, protein) in a sequence record. It contains information such as the feature
        type, location, qualifiers (e.g., product, locus_tag, external_id),

        Returns:
          a dictionary
        """
        note = {}
        if "note" in seq_feature.qualifiers:
            temp_note = ";".join(seq_feature.qualifiers["note"])
            note = note | self._process_feature_note_protein_exceptions(temp_note)
            note = note | self._process_go(temp_note)
        return note

    def _add_feature(
        self,
        seq_record,
        seq_feature,
        assembly_id,
        locus_id,
        keep_sequence,
    ):
        """
        Adds a sequence feature to the child SocialGene object

        Args:
          seq_record: The `seq_record` parameter is a BioPython `SeqRecord` object, which represents a
        biological sequence record. It contains information about the sequence, such as its sequence
        data, annotations, and features.
          seq_feature: The `seq_feature` parameter is a SeqFeature object that represents a feature
        (e.g., gene, CDS, protein) in a sequence record. It contains information such as the feature
        type, location, qualifiers (e.g., product, locus_tag, external_id)
          assembly_id: The `assembly_id` parameter is used to identify the assembly to which the
        sequence feature belongs.
          locus_id: The `locus_id` parameter is used to identify a specific locus.
          keep_sequence: The `keep_sequence` parameter is a boolean flag that determines whether the
        sequence of the added feature should be retained or not.
        """
        # Grab the locus/protein description if it exists
        if "product" in seq_feature.qualifiers:
            description = ";".join(seq_feature.qualifiers["product"]).strip()
        else:
            description = None
        if seq_feature.type == "source":
            # TODO: should this overwrite (current) or compare every iteration per assembly?
            self.assemblies[assembly_id].metadata.update(seq_feature.qualifiers)
            self.assemblies[assembly_id].loci[locus_id].metadata.update(
                seq_feature.qualifiers
            )
        if seq_feature.type in ["protein", "CDS"]:
            try:
                locus_tag = None
                external_id = None
                if "locus_tag" in seq_feature.qualifiers:
                    locus_tag = ";".join(seq_feature.qualifiers["locus_tag"])

                if "protein_id" in seq_feature.qualifiers:
                    external_id = ";".join(seq_feature.qualifiers["protein_id"])
                elif locus_tag:
                    # use locus tag as protein instead
                    external_id = locus_tag
                else:
                    # else assign random id
                    external_id = str(uuid4())
                if "translation" in seq_feature.qualifiers:
                    translation = seq_feature.qualifiers["translation"][0]
                elif "pseudo" or "pseudogene" in seq_feature.qualifiers:
                    # removes biopython potential deprecation warning to prepend Ns if non-modulo 3 seqs,
                    # may be brittle if biopython changes
                    translation = str(
                        Seq.Seq(seq_feature.extract(seq_record).seq).translate()
                    )
                    description = f"pseudo_{description}"
                else:
                    raise ValueError(
                        f"Panic!!! Not a protein or pseudo protein: {seq_feature.qualifiers}"
                    )
                uid = self.add_protein(
                    description=description,
                    external_id=external_id.strip(),
                    sequence=translation.strip(),
                    return_uid=True,
                )
                if not keep_sequence:
                    self.proteins[uid].sequence = None
                self.assemblies[assembly_id].loci[locus_id].add_feature(
                    type=seq_feature.type,
                    uid=uid,
                    external_id=external_id.strip(),
                    start=get_seqio_start(seq_feature),
                    end=get_seqio_end(seq_feature),
                    strand=seq_feature.location.strand,
                    locus_tag=locus_tag,
                    description=description,
                    **self._process_feature_note(seq_feature),
                )

            except Exception as e:
                raise e

    def _parse_record(self, seq_record, assembly_id, keep_sequence):
        """
        Parses a sequence record, adds a locus to the record, adds features
        to the record, and updates the count of loci in the file.

        Args:
          seq_record: The `seq_record` parameter is an instance of the `SeqRecord` class, which
        represents a biological sequence record. It contains information about the sequence, such as its
        sequence data, annotations, and features.
          assembly_id: The `assembly_id` parameter is an identifier for the assembly to which the
        sequence record belongs. It is used to associate the sequence record and its features with a
        specific assembly.
          keep_sequence: The parameter "keep_sequence" is a boolean value that determines whether or not
        to keep the sequence information when adding a feature.
        """
        locus_id = self._add_locus(seq_record=seq_record, assembly_id=assembly_id)
        for seq_feature in seq_record.features:
            self._add_feature(
                seq_record=seq_record,
                seq_feature=seq_feature,
                assembly_id=assembly_id,
                locus_id=locus_id,
                keep_sequence=keep_sequence,
            )
        self._count_loci_in_file = self._count_loci_in_file + Counter(
            [i.type for i in seq_record.features]
        )

    def _parse_genbank(
        self,
        handle,
        input_path,
        assembly_id=None,
        keep_sequence=True,
        **kwargs,
    ):
        """Internal function that parses a genbank file, adding assembly and taxon information,
         and then calls another function to parse the record and its features.
        Args:
            handle (StringIO): genbank file handle
            input_path (str): Path to genbank file (filename used as assembly name, if missing)
            assembly_id (str, optional): Assign the assembly id (used for for neo4j)
            keep_sequence (bool, optional): Store the amino acid sequence of proteins?
        """
        # only set assembly info on reading first record
        stop_after_one = True
        for seq_record in SeqIO.parse(handle, "genbank"):
            # Add assembly
            if stop_after_one:
                if not assembly_id:
                    assembly_id = self._add_assembly(
                        input_path=input_path, seq_record=seq_record
                    )
                self._add_taxon(
                    seq_feature=seq_record.features[0], assembly_id=assembly_id
                )
                stop_after_one = False
            # Add locus and its features
            log.debug(f"Processing assembly: {assembly_id}")
            self._parse_record(
                seq_record=seq_record,
                assembly_id=assembly_id,
                keep_sequence=keep_sequence,
            )
        log.debug(f"Processing seq_record: {seq_record.id}")

    def _open_genbank(self, input_path, **kwargs):
        """
        The `_open_genbank` function is used to parse a genbank file, which can handle .gz and tar
        archives.

        Args:
          input_path: The `input_path` parameter is a string that represents the path to the genbank
        file that you want to parse. It should be able to handle both uncompressed files and files
        compressed with gzip or tar. (tar hasn't been tested in a while)
        """
        # self._count_loci_in_file is mostly for sending a summary of parsed features to the logger
        log.info(f"Parsing: {input_path}")
        self._count_loci_in_file = Counter()
        input_path = Path(input_path)
        if fh.check_if_tar(input_path):
            tar = fh.opener_result(input_path)
            log.info(f"tar archive has {len(tar.members)} files")
            log.info(tar)
            for member in tar:
                f = tar.extractfile(member)
                if "gbk" in member.name or "gbff" in member.name:
                    try:
                        log.info(f"Extracting and parsing: {member.name}")
                        input_ = zlib.decompress(f.read(), 16 + zlib.MAX_WBITS).decode(
                            "utf-8"
                        )
                        with StringIO(input_) as handle:
                            self._parse_genbank(
                                handle=handle,
                                input_path=input_path,
                            )
                    except Exception as e:
                        log.debug(e)
        else:
            with fh.open_read(input_path) as handle:
                self._parse_genbank(
                    handle=handle,
                    input_path=input_path,
                    **kwargs,
                )
        log.info(
            f"'{input_path}' features {dict(sorted(dict(self._count_loci_in_file).items(), key=lambda item: item[1], reverse=True))}"
        )

    @staticmethod
    def _extract_from_dbxrefs(input_list, prefix):
        """Helper function for genbank parser, extract a prefixed string from list

        Args:
            input_list (list): e.g. ['BioProject:PRJNA224116', 'BioSample:SAMN00006196', 'Assembly:GCF_000145235.1']
            prefix (str): e.g. "BioProject:"

        Returns:
            list: "PRJNA224116"
        """
        temp = [x.removeprefix(prefix) for x in input_list if x.startswith(prefix)]
        return ";".join(temp)



import argparse

from socialgene.neo4j.utils.admin_import import Neo4jAdminImport

parser = argparse.ArgumentParser(
    description="Write the header files for neo4j admin import"
)

parser.add_argument(
    "--neo4j_top_dir",
    help="Output directory where header files will be written to",
    required=True,
)

parser.add_argument(
    "--sg_modules",
    help="List of relationship labels to not create arguments for; e.g. to add mmseqs: base,mmseqs2 ",
    required=True,
    default="base",
    nargs="+",
)

parser.add_argument(
    "--cpus",
    help="Number of cpus to give to neo4j admin import",
    required=False,
    default=4,
)

parser.add_argument(
    "--docker_version",
    help="Sets 'chasemc2/sgnf-sgpy:{docker_version}'",
    required=False,
    default="latest",
)

parser.add_argument(
    "--additional_args",
    help="Additional arguments to pass to the neo4j admin import docker command, given as single string",
    required=False,
    default="",
)
parser.add_argument(
    "--uid",
    type=int,
    help="uid to give to docker, default is the current user",
    required=False,
)
parser.add_argument(
    "--gid",
    type=int,
    help="gid to give to docker, default is the current user",
    required=False,
)
parser.add_argument(
    "--dryrun",
    help="just write to file instead of running",
    required=False,
    default=None,
    action=argparse.BooleanOptionalAction,
)
parser.add_argument(
    "--dryrun_filepath",
    help="Filepath of neo4j import script that will be created",
    required=False,
    default="./build_db.sh",
)
parser.add_argument(
    "--docker",
    help="Use this flag to to add docker-specific commands",
    required=False,
    default=False,
    action=argparse.BooleanOptionalAction,
)


def main():
    args = parser.parse_args()

    if args.uid == "None":
        uid = None
    else:
        uid = args.uid

    if args.gid == "None":
        gid = None
    else:
        gid = args.gid

    nai_obj = Neo4jAdminImport(
        neo4j_top_dir=args.neo4j_top_dir,
        module_list=args.sg_modules,
        cpus=int(args.cpus),
        additional_args=args.additional_args,
        uid=uid,
        gid=gid,
    )
    tmp = nai_obj.run(docker=args.docker, dryrun=args.dryrun)
    if args.dryrun:
        with open(args.dryrun_filepath, "w") as handle:
            handle.writelines(" \\\n\t".join(tmp))


if __name__ == "__main__":
    main()



import csv
from pathlib import Path

import socialgene.utils.file_handling as fh
from socialgene.parsers.hmmmodel import HmmModel


class HmmInfo:
    def __init__(self, all_hmms_path):
        """
        The function initializes an object with a list of columns and a path to a file.

        Args:
          all_hmms_path: The `all_hmms_path` parameter is a string that represents the path to a file
        containing data for all HMMs.
        """
        self.columns = list(HmmModel().all_attributes().keys())
        self.all_hmms_path = all_hmms_path
        self.all_hmms_data = list()

    def read_tsv(self):
        """
        The function reads a TSV file, validates the number of columns, and appends the data to a list.
        """
        with fh.open_read(self.all_hmms_path) as h:
            for line in h:
                line_vals = [
                    None if v.strip() == '""' else v.strip() for v in line.split("\t")
                ]
                if not len(line_vals) == 16:
                    raise ValueError(f"Expected 16 columns, not {len(line_vals)}")
                if not line_vals == self.columns:
                    # Check for header, if exists, skip reading it in
                    self.all_hmms_data.append(tuple(line_vals))

    def write_nr_hmm_nodes(self, outpath="sg_hmm_nodes"):
        """This writes the HMM node file

        Args:
            single_line_list (list): list representing one row of the HMM metadata file
        """
        with open(outpath, "w") as handle:
            for i in list({i[9] for i in self.all_hmms_data}):
                handle.write(i)
                handle.write("\n")

    def write_hmm_source_nodes(self, outdir="."):
        """Split the HMM metadata into parts based on the database it comes from

        Args:
            single_line_list (list): list representing one row of the HMM metadata file
        """

        for hmm_database_name in {i[0] for i in self.all_hmms_data}:
            # "_hmm_source" is to help glob these files within nextflow
            outpath = Path(outdir, f"{hmm_database_name}_hmm_source")
            with open(outpath, "w") as handle:
                tsv_writer = csv.writer(
                    handle, delimiter="\t", quotechar='"', quoting=csv.QUOTE_MINIMAL
                )
                for i in (i for i in self.all_hmms_data if i[0] == hmm_database_name):
                    tsv_writer.writerow(i[1:])


# write a context manager class that takes a string and returns the string
class PassThrough:
    def __init__(self, string):
        self.string = string

    def __enter__(self):
        return self.string

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass



from socialgene.neo4j.neo4j_element import Node, Relationship


class ChemicalCompoundNode(Node):
    neo4j_label = ["chemical_compound"]
    description = "Represents a chemical compound"
    required_properties = ["inchi", "CanonicalSmiles"]
    property_specification = {
        "uid": str,
        "MolWt": float,
        "HeavyAtomMolWt": float,
        "ExactMolWt": float,
        "NumValenceElectrons": int,
        "NumRadicalElectrons": int,
        "HeavyAtomCount": int,
        "NumAliphaticCarbocycles": int,
        "NumAliphaticHeterocycles": int,
        "NumAliphaticRings": int,
        "NumAromaticCarbocycles": int,
        "NumAromaticHeterocycles": int,
        "NumAromaticRings": int,
        "NumHAcceptors": int,
        "NumHDonors": int,
        "NumHeteroatoms": int,
        "NumRotatableBonds": int,
        "NumSaturatedCarbocycles": int,
        "NumSaturatedHeterocycles": int,
        "NumSaturatedRings": int,
        "RingCount": int,
        "MolLogP": float,
        "MolMR": float,
        "AnonymousGraph": str,
        "ElementGraph": str,
        "MurckoScaffold": str,
        "ExtendedMurcko": str,
        "MolFormula": str,
        "AtomBondCounts": str,
        "DegreeVector": str,
        "Mesomer": str,
        "HetAtomTautomer": str,
        "HetAtomProtomer": str,
        "RedoxPair": str,
        "Regioisomer": str,
        "NetCharge": str,
        "SmallWorldIndexBR": str,
        "SmallWorldIndexBRL": str,
        "ArthorSubstructureOrder": str,
        "HetAtomTautomerv2": str,
        "inchi": str,
        "CanonicalSmiles": str,
    }
    constraints_unique = ["inchi"]


class TanimotoSimilarity(Relationship):
    neo4j_label = "TANIMOTO_SIMILARITY"
    description = "Connects two chemical compounds that are similar"
    start_class = ChemicalCompoundNode
    end_class = ChemicalCompoundNode
    property_specification = {
        "similarity": int,
    }


class McsSimilarity(Relationship):
    neo4j_label = "MCS_SIMILARITY"
    description = "Connects two chemical compounds that have similar maximium common substructures"
    start_class = ChemicalCompoundNode
    end_class = ChemicalCompoundNode
    property_specification = {
        "similarity": int,
    }


class ChemicalSubstructure(ChemicalCompoundNode):
    neo4j_label = ["substructure"]
    description = "Represents a chemical substructure"
    required_properties = ["inchi", "CanonicalSmiles"]
    property_specification = {
        "uid": str,
        "inchi": str,
        "CanonicalSmiles": str,
    }
    constraints_unique = ["inchi", "CanonicalSmiles"]


class ContainsSubstructure(Relationship):
    neo4j_label = "SUBSTRUCTURE"
    description = "Connects a chemical compound to a chemical substructure"
    start_class = ChemicalCompoundNode
    end_class = ChemicalSubstructure



"""https://www.npatlas.org"""

from socialgene.addons.chebi.nr import ChebiNode
from socialgene.addons.chemistry.nr import ChemicalCompoundNode
from socialgene.addons.classyfire.nr import ClassyFireNode
from socialgene.addons.gnps_library.nr import GnpsLibrarySpectrumNode
from socialgene.addons.mibig.nr import Mibig
from socialgene.addons.npclassifier.nr import (
    NPClassifierClass,
    NPClassifierPathway,
    NPClassifierSuperclass,
)
from socialgene.addons.npmrd.nr import Npmrd
from socialgene.addons.publication.nr import Publication
from socialgene.neo4j.neo4j_element import Node, Relationship
from socialgene.nextflow.nodes import TAXID


class NPAtlasNode(Node):
    neo4j_label = ["npatlas_compound"]
    description = "Represents a single NPAtlas entry"
    property_specification = {
        "uid": str,
        "original_name": str,
        "mol_formula": str,
        "mol_weight": float,
        "exact_mass": float,
        "inchikey": str,
        "smiles": str,
        "cluster_id": int,
        "node_id": int,
        "synonyms": str,
        "inchi": str,
        "m_plus_h": float,
        "m_plus_na": float,
        "genus": str,
        "species": str,
    }
    required_properties = ["uid"]
    constraints_unique = ["uid"]


class MibigToNPAtlas(Relationship):
    neo4j_label = "PRODUCES"
    description = "Connects an NPAtlas entry to a Mibig entry"
    start_class = Mibig
    end_class = NPAtlasNode


class NPAtlasToNpmrd(Relationship):
    neo4j_label = "HAS"
    description = "Connects an NPAtlas entry to an Npmrd entry"
    start_class = NPAtlasNode
    end_class = Npmrd


class NPAtlasToGnps(Relationship):
    neo4j_label = "HAS"
    description = "Connects an NPAtlas entry to a GNPS entry"
    start_class = NPAtlasNode
    end_class = GnpsLibrarySpectrumNode


class NPAtlasPublication(Publication):
    pass


class NPAtlasToPublication(Relationship):
    neo4j_label = "HAS"
    description = "Connects an NPAtlas entry to a publication"
    start_class = NPAtlasNode
    end_class = NPAtlasPublication


class NPAtlasToTaxID(Relationship):
    neo4j_label = "PRODUCES"
    description = "Connects an NPAtlas entry to a TaxID entry"
    start_class = TAXID
    end_class = NPAtlasNode


class NPAtlasToClassyFireDirectParent(Relationship):
    neo4j_label = "DIRECT_PARENT"
    description = "Connects an NPAtlas entry to a ClassyFire entry"
    start_class = NPAtlasNode
    end_class = ClassyFireNode


class NPAtlasToClassyFireLowestClass(Relationship):
    neo4j_label = "LOWEST_CLASS"
    description = "Connects an NPAtlas entry to a ClassyFire entry"
    start_class = NPAtlasNode
    end_class = ClassyFireNode


class NPAtlasToClassyFireIntermediateNodes(Relationship):
    neo4j_label = "INTERMEDIATE_NODES"
    description = "Connects an NPAtlas entry to a ClassyFire entry"
    start_class = NPAtlasNode
    end_class = ClassyFireNode


class NPAtlasToClassyFireAlternativeParents(Relationship):
    neo4j_label = "ALTERNATIVE_PARENTS"
    description = "Connects an NPAtlas entry to a ClassyFire entry"
    start_class = NPAtlasNode
    end_class = ClassyFireNode


class NPAtlasToNpclassifierClass(Relationship):
    neo4j_label = "IS_A"
    description = "Connects an NPAtlas entry to a NPClassifierClass entry"
    start_class = NPAtlasNode
    end_class = NPClassifierClass


class NPAtlasToNpclassifierPathway(Relationship):
    neo4j_label = "IS_A"
    description = "Connects an NPAtlas entry to a NPClassifierPathway entry"
    start_class = NPAtlasNode
    end_class = NPClassifierPathway


class NPAtlasToNpclassifierSuperclass(Relationship):
    neo4j_label = "IS_A"
    description = "Connects an NPAtlas entry to a NPClassifierSuperclass entry"
    start_class = NPAtlasNode
    end_class = NPClassifierSuperclass


class NPAtlasToChem(Relationship):
    neo4j_label = "IS_A"
    description = "Connects an NPAtlas entry to a ChemicalCompound entry"
    start_class = NPAtlasNode
    end_class = ChemicalCompoundNode


class NPAtlasToChebi(Relationship):
    neo4j_label = "IS_A"
    description = "Connects an NPAtlas entry to a ChEBI entry"
    start_class = NPAtlasNode
    end_class = ChebiNode
